# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pwem.protocols import EMProtocol
import pyworkflow.protocol.params as params
from pwem.objects.data import AtomStruct
from .utilities.pdb_handler import ContinuousFlexPDBHandler
from pyworkflow.utils import runCommand
import os
from pwem.convert.atom_struct import cifToPdb


NUCLEIC_NO = 0
NUCLEIC_RNA =1
NUCLEIC_DNA = 2

FORCEFIELD_CHARMM = 0
FORCEFIELD_AAGO = 1
FORCEFIELD_CAGO = 2


class ProtGenerateTopology(EMProtocol):
    """ Protocol to generate topology files for GENESIS simulations """
    _label = 'generate topology'

    def _defineParams(self, form):

        form.addSection(label='Inputs')

        form.addParam('inputPDB', params.PointerParam,
                      pointerClass='AtomStruct', label="Input PDB",
                      help='Select the input PDB.', important=True)

        group = form.addGroup('Forcefield Inputs')
        group.addParam('forcefield', params.EnumParam, label="Forcefield type", default=FORCEFIELD_CHARMM, important=True,
                       choices=['CHARMM', 'All-atom Go model', 'C-Alpha Go model'],
                       help="Type of the force field used for energy and force calculation. For Go models, it is strongly"
                            " recommended to first generate topology using CHARMM, then create a new protocol to generate"
                            " Go model topology based on the output CHARMM all-atom PDB model."
                            " This will ensure that residue sequences are consecutive and TER statements are present in PDB."
                            " CHARMM requires VMD psfgen installed. Go models requires SMOG 2 installed. ")
        group.addParam('nucleicChoice', params.EnumParam, label="Contains nucleic acids ?", default=NUCLEIC_NO,
                       choices=['No', 'RNA', 'DNA'], help="Specify if the generator should consider nucleic residues as DNA or RNA")

        group.addParam('inputPRM', params.FileParam, label="CHARMM parameter file (prm)",
                       condition="forcefield==%i"%FORCEFIELD_CHARMM,
                       help='CHARMM parameter file containing force field parameters, e.g. force constants and librium'
                            ' geometries. Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ')
        group.addParam('inputRTF', params.FileParam, label="CHARMM topology file (rtf)",
                       condition="forcefield==%i"%FORCEFIELD_CHARMM,
                       help='CHARMM topology file containing information about atom connectivity of residues and'
                            ' other molecules. Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ')
        group.addParam('inputSTR', params.FileParam, label="CHARMM stream file (str, optional)",
                       condition="forcefield==%i"%FORCEFIELD_CHARMM, default="",
                       help='CHARMM stream file containing both topology information and parameters. '
                            'Latest forcefields can be founded at http://mackerell.umaryland.edu/charmm_ff.shtml ')

        group.addParam('smog_dir', params.FileParam, label="SMOG 2 install directory",
                       help="Path to SMOG2 install directory (For SMOG2 installation, see "
                            "https://smog-server.org/smog2/). If SMOG2 is not installed, you can use the web GUI instead "
                            "https://smog-server.org/cgi-bin/GenTopGro.pl (Recommended to run the protocol with empty smog_dir,"
                            " when the protocol fails, get the input.pdb file generate in the extra directory as input of SMOG server)",
                       condition="(forcefield==%i or forcefield==%i)"%(FORCEFIELD_CAGO, FORCEFIELD_AAGO))

    def _insertAllSteps(self):
        ff = self.forcefield.get()

        self._insertFunctionStep("convertInput")

        if ff == FORCEFIELD_CAGO or ff == FORCEFIELD_AAGO:
            self._insertFunctionStep("prepareGROTOP")
            self._insertFunctionStep("runGROTOP")

        if ff == FORCEFIELD_CHARMM:
            self._insertFunctionStep("preparePSF")
            self._insertFunctionStep("runPSF")

        self._insertFunctionStep("checkPDB")
        self._insertFunctionStep("createOutput")

    def convertInput(self):
        inputPDB = self.inputPDB.get().getFileName()
        outPDB = self._getExtraPath("input.pdb")
        ext = os.path.splitext(inputPDB)[1]

        if ext == ".pdb" or ext == ".ent" :
            runCommand("cp %s %s" % (inputPDB, outPDB))
        elif ext == ".cif" or ext == ".mmcif" :
            cifToPdb(inputPDB, outPDB)
        else:
            print("ERROR (toPdb), Unknown file type for file = %s" % inputPDB)

    def createOutput(self):
        self._defineOutputs(outputPDB=AtomStruct(self._getExtraPath("output.pdb")))

    def preparePSF(self):
        inputPDB = self._getExtraPath("input.pdb")
        inputTopo = self.inputRTF.get()
        outputPrefix = self._getExtraPath("output")
        nucleicChoice = self.nucleicChoice.get()

        fnPSFgen = self._getExtraPath("psfgen.tcl")
        with open(fnPSFgen, "w") as psfgen:
            psfgen.write("mol load pdb %s\n" % inputPDB)
            psfgen.write("\n")
            psfgen.write("package require psfgen\n")
            psfgen.write("topology %s\n" % inputTopo)
            psfgen.write("pdbalias residue HIS HSE\n")
            psfgen.write("pdbalias residue MSE MET\n")
            psfgen.write("pdbalias atom ILE CD1 CD\n")
            if nucleicChoice == NUCLEIC_RNA:
                psfgen.write("pdbalias residue A ADE\n")
                psfgen.write("pdbalias residue G GUA\n")
                psfgen.write("pdbalias residue C CYT\n")
                psfgen.write("pdbalias residue U URA\n")
            elif nucleicChoice == NUCLEIC_DNA:
                psfgen.write("pdbalias residue DA ADE\n")
                psfgen.write("pdbalias residue DG GUA\n")
                psfgen.write("pdbalias residue DC CYT\n")
                psfgen.write("pdbalias residue DT THY\n")
            psfgen.write("\n")
            if nucleicChoice == NUCLEIC_RNA or nucleicChoice == NUCLEIC_DNA:
                psfgen.write("set nucleic [atomselect top nucleic]\n")
                psfgen.write("set chains [lsort -unique [$nucleic get chain]] ;\n")
                psfgen.write("foreach chain $chains {\n")
                psfgen.write("    set sel [atomselect top \"nucleic and chain $chain\"]\n")
                psfgen.write("    $sel writepdb %s_tmp.pdb\n" % outputPrefix)
                psfgen.write("    segment N${chain} { pdb %s_tmp.pdb }\n" % outputPrefix)
                psfgen.write("    coordpdb %s_tmp.pdb N${chain}\n" % outputPrefix)
                if nucleicChoice == NUCLEIC_DNA:
                    psfgen.write("    set resids [lsort -unique [$sel get resid]]\n")
                    psfgen.write("    foreach r $resids {\n")
                    psfgen.write("        patch DEOX N${chain}:$r\n")
                    psfgen.write("    }\n")
                psfgen.write("}\n")
                if nucleicChoice == NUCLEIC_DNA:
                    psfgen.write("regenerate angles dihedrals\n")
                psfgen.write("\n")
            psfgen.write("set protein [atomselect top protein]\n")
            psfgen.write("set chains [lsort -unique [$protein get pfrag]]\n")
            psfgen.write("foreach chain $chains {\n")
            psfgen.write("    set sel [atomselect top \"protein and pfrag $chain\"]\n")
            psfgen.write("    $sel writepdb %s_tmp.pdb\n" % outputPrefix)
            psfgen.write("    segment P${chain} {pdb %s_tmp.pdb}\n" % outputPrefix)
            psfgen.write("    coordpdb %s_tmp.pdb P${chain}\n" % outputPrefix)
            psfgen.write("}\n")
            psfgen.write("rm -f %s_tmp.pdb\n" % outputPrefix)
            psfgen.write("\n")
            psfgen.write("guesscoord\n")
            psfgen.write("writepdb %s.pdb\n" % outputPrefix)
            psfgen.write("writepsf %s.psf\n" % outputPrefix)
            psfgen.write("exit\n")

    def checkPDB(self):
        outPDB = self._getExtraPath("output.pdb")

        # Check PDB
        if not os.path.isfile(outPDB) :
            raise RuntimeError("Can not locate output PDB file %s, check log files for more details " % outPDB)
        if os.path.getsize(outPDB) ==0 :
            raise RuntimeError("PDB file %s is empty, check log files for more details " % outPDB)

        outMol = ContinuousFlexPDBHandler(outPDB)
        if outMol.n_atoms == 0:
            raise RuntimeError("PDB file %s is empty, check log files for more details " % outPDB)

    def runPSF(self):
        fnPSFgen = self._getExtraPath("psfgen.tcl")
        outputPrefix = self._getExtraPath("output")

        # Run VMD PSFGEN
        runCommand("vmd -dispdev text -e %s > %s.log " % (fnPSFgen, outputPrefix))


    def prepareGROTOP(self):
        inputPDB = self._getExtraPath("input.pdb")

        mol = ContinuousFlexPDBHandler(inputPDB)
        # mol.remove_alter_atom()
        mol.remove_hydrogens()
        mol.check_res_order()

        mol.alias_atom("CD", "CD1", "ILE")
        mol.alias_atom("OT1", "O")
        mol.alias_atom("OT2", "OXT")
        mol.alias_res("HSE", "HIS")
        mol.alias_res("HSD", "HIS")
        mol.alias_res("HSP", "HIS")

        if self.nucleicChoice.get() == NUCLEIC_RNA:
            mol.alias_res("CYT", "C")
            mol.alias_res("GUA", "G")
            mol.alias_res("ADE", "A")
            mol.alias_res("URA", "U")

        elif self.nucleicChoice.get() == NUCLEIC_DNA:
            mol.alias_res("CYT", "DC")
            mol.alias_res("GUA", "DG")
            mol.alias_res("ADE", "DA")
            mol.alias_res("THY", "DT")

        mol.alias_atom("O1'", "O1*")
        mol.alias_atom("O2'", "O2*")
        mol.alias_atom("O3'", "O3*")
        mol.alias_atom("O4'", "O4*")
        mol.alias_atom("O5'", "O5*")
        mol.alias_atom("C1'", "C1*")
        mol.alias_atom("C2'", "C2*")
        mol.alias_atom("C3'", "C3*")
        mol.alias_atom("C4'", "C4*")
        mol.alias_atom("C5'", "C5*")
        mol.alias_atom("C5M", "C7")
        mol.add_terminal_res()
        mol.atom_res_reorder()
        mol.write_pdb(inputPDB)

    def runGROTOP(self):
        outputPrefix = self._getExtraPath("output")
        inputPDB = self._getExtraPath("input.pdb")

        # Run Smog2
        runCommand("%s/bin/smog2" % self.smog_dir.get() + \
                   " -i %s -dname %s -%s -limitbondlength -limitcontactlength > %s.log" %
                   (inputPDB, outputPrefix,
                    "CA" if self.forcefield.get() == FORCEFIELD_CAGO else "AA", outputPrefix))

        # ADD CHARGE TO TOP FILE
        grotopFile = outputPrefix + ".top"
        with open(grotopFile, 'r') as f1:
            with open(grotopFile + ".tmp", 'w') as f2:
                atom_scope = False
                write_line = False
                for line in f1:
                    if "[" in line and "]" in line:
                        if "atoms" in line:
                            atom_scope = True
                    if atom_scope:
                        if "[" in line and "]" in line:
                            if not "atoms" in line:
                                atom_scope = False
                                write_line = False
                        elif not ";" in line and not (not line or line.isspace()):
                            write_line = True
                        else:
                            write_line = False
                    if write_line:
                        f2.write("%s\t0.0\n" % line[:-1])
                    else:
                        f2.write(line)
        runCommand("cp %s.tmp %s" % (grotopFile, grotopFile))
        runCommand("rm -f %s.tmp" % grotopFile)

        if self.forcefield.get() == FORCEFIELD_CAGO:
            mol = ContinuousFlexPDBHandler(inputPDB)
            mol.select_atoms(mol.allatoms2ca())
            mol.write_pdb(outputPrefix + ".pdb")
        else:
            runCommand("cp %s %s"%(inputPDB,outputPrefix + ".pdb"))

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _citations(self):
        return ['harastani2022continuousflex','vuillemot2022NMMD']

    def _methods(self):
        pass



# **************************************************************************
# * Authors:    Ilyes Hamitouche            (ilyes.hamitouche@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from continuousflex.protocols.protocol_deep_hemnma_train import FlexProtDeepHEMNMATrain
from pyworkflow.protocol.params import LabelParam
from pyworkflow.viewer import ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO
from subprocess import check_call
import sys
from continuousflex import Plugin


class FlexDeepHEMNMAViewer(ProtocolViewer):
    """ Visualization of results from the deepHEMNMA training protocol
    """
    _label = 'viewer deepHEMNMA'
    _targets = [FlexProtDeepHEMNMATrain]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self._data = None

    def _defineParams(self, form):
        form.addSection(label='Visualization')
        form.addParam('displaycurves', LabelParam,
                      label="Display training curves",
                      help="Display the training and validation losses")


    def _getVisualizeDict(self):
        return {'displaycurves': self._viewcurves}


    def _viewcurves(self, pramName):
        import webbrowser
        logdir = self.protocol._getExtraPath('scalars/')
        command = "tensorboard --port=6006 --logdir " + logdir +'&'
        command = Plugin.getContinuousFlexCmd(command)
        check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr, env=None, cwd=None)
        webbrowser.open_new("http://localhost:6006/")



# **************************************************************************
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import joblib
from pwem.protocols import ProtAnalysis3D
import xmipp3.convert
import pwem.emlib.metadata as md
import pyworkflow.protocol.params as params
from pyworkflow.utils.path import makePath, createLink
from continuousflex.protocols.utilities.spider_files3 import save_volume
import sys
from pyworkflow.utils import getListFromRangeString
from os.path import isfile
from joblib import Parallel, delayed
import continuousflex
from subprocess import check_call
from pwem.utils import runProgram
from pwem.emlib.image import ImageHandler
import numpy as np
from continuousflex import Plugin

REFERENCE_EXT = 0
REFERENCE_STA = 1

IMPORT_FLOWS = 0
FIND_FLOWS = 1

class FlexProtHeteroFlow(ProtAnalysis3D):
    """ Protocol for TomoFlow. """
    _label = 'tomoflow protocol'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Settings')
        group = form.addGroup('Choose what processes you want to perform:')
        group.addParam('copy_opflows', params.EnumParam,
                      choices=['Import optical flows from the last refinement iteration and analyze the heterogeneity',
                               'Find optical flows for a set of ALIGNED volumes and analyze the heterogeneity'],
                      default=IMPORT_FLOWS,
                      label='Optical flows to analyze', display=params.EnumParam.DISPLAY_COMBO,
                       help='You can choose to find the optical flows for a set of volumes or to import'
                            ' precalculated optical flows from a refinement protocol previous run in the project'
                            ' workspace')
        group = form.addGroup('Protocol of: Missing wedge correction and combined rigid-body and elastic alignment',
                              condition='copy_opflows==%d'% IMPORT_FLOWS)
        group.addParam('refinementProt', params.PointerParam, pointerClass='FlexProtRefineSubtomoAlign',
                       label='Point to the refinement protocol', allowsNull=True)

        group = form.addGroup('Input', condition='copy_opflows==%d' % FIND_FLOWS)
        group.addParam('inputVolumes', params.PointerParam,
                      pointerClass='SetOfVolumes', allowsNull=True,
                      label="Input volume(s)", important=True,
                      help='Select volumes')
        group.addParam('StartingReference', params.EnumParam,
                      choices=['From an external volume file', 'Select a volume'],
                      default=REFERENCE_STA,
                      label='Reference volume', display=params.EnumParam.DISPLAY_COMBO,
                      help='Either an external volume file or a subtomogram average')
        group.addParam('ReferenceVolume', params.FileParam,
                      pointerClass='params.FileParam', allowsNull=True,
                      condition='StartingReference==%d' % REFERENCE_EXT,
                      label="Reference volume",
                      help='Choose a reference, typically from a STA previous run')
        group.addParam('STAVolume', params.PointerParam,
                      pointerClass='Volume', allowsNull=True,
                      condition='StartingReference==%d' % REFERENCE_STA,
                      label="Selected volume",
                      help='Choose a reference, typically from a STA previous run')
        group.addParam('WarpAndEstimate', params.BooleanParam,
                      expertLevel=params.LEVEL_ADVANCED,
                      default=True,
                      label='Save a warped version of the reference for each input volume?',
                      help='This will create a set of volumes, representing the fitted version of the input volumes, '
                           'using the calculated optical flows, and calculate the cross correlation, mean square '
                           'distance and the mean absolute distance between the input volumes and estimated volumes')
        form.addSection(label='3D OpticalFLow parameters')
        group = form.addGroup('Optical flows', condition='copy_opflows==%d' % FIND_FLOWS)
        group.addParam('N_GPU', params.IntParam, default=1, important=True, allowsNull=True,
                              label = 'Parallel processes on GPU',
                              help='This parameter indicates the number of volumes that will be processed in parallel'
                                   ' (independently). The more powerful your GPU, the higher the number you can choose.')
        group.addParam('GPU_list', params.NumericRangeParam,
                       label="GPU id(s)",
                       help='Select the GPU id(s) that will be used for optical flow calculation.'
                            'Examples: '
                            'You can select a list like 0-4, and it will take the GPUs 0 1 2 3 4'
                            'You can also combine different selections like 1, 3-5 and it will take 1, 3, 4, 5',
                       default='0')
        group.addParam('pyr_scale', params.FloatParam, default=0.5,
                      label='pyr_scale', allowsNull=True,
                       help='parameter specifying the image scale to build pyramids for each image (pyr_scale < 1). '
                            'A classic pyramid is of generally 0.5 scale, every new layer added, it is halved to the previous one.')
        group.addParam('levels', params.IntParam, default=4, allowsNull=True,
                      label='levels',
                      help='levels=1 says, there are no extra layers (only the initial image).'
                           ' It is the number of pyramid layers including the first image.'
                           ' The coarsest possible pyramid level is 32x32x32 voxels')
        group.addParam('winsize', params.IntParam, default=10, allowsNull=True,
                      label='winsize',
                      help='It is the averaging window size, larger the size, the more robust the algorithm is to noise, '
                           'and provide fast motion detection, though gives blurred motion fields.')
        group.addParam('iterations', params.IntParam, default=10, allowsNull=True,
                      label='iterations',
                      help='Number of iterations to be performed at each pyramid level. It refines the optical flow'
                           ' accuracy at each scale and allows accounting for larger displacements.')
        group.addParam('poly_n', params.IntParam, default=5, allowsNull=True,
                      label='poly_n',
                      help='Size of the pixel neighborhood used to find polynomial expansion in each pixel;'
                           ' larger values mean that the image will be approximated with smoother surfaces,'
                           ' yielding more robust algorithm and more blurred motion field, typically poly_n = 5 or 7.')
        group.addParam('poly_sigma', params.FloatParam, default=1.2,
                      label='poly_sigma',
                      help='Standard deviation of the Gaussian that is used to smooth derivatives used as '
                           'a basis for the polynomial expansion; for poly_n = 5, you can set poly_sigma = 1.2,'
                           ' for poly_n = 7, a good value would be poly_sigma = 1.5.')
        group.addHidden('flags', params.IntParam, default=0,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='flags',
                      help='flag to pass for the optical flow')
        group.addHidden('factor1', params.IntParam, default=100,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='gray scale factor1',
                      help='this factor will be multiplied by the gray levels of each subtomogram')
        group.addHidden('factor2', params.IntParam, default=100,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='gray scale factor2',
                      help='this factor will be multiplied by the gray levels of the reference')

    # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        # Define some outputs filenames
        self.imgsFn = self._getExtraPath('volumes.xmd')

        if(self.copy_opflows.get()==FIND_FLOWS):
            makePath(self._getExtraPath() + '/optical_flows')
            self._insertFunctionStep('convertInputStep')
            self._insertFunctionStep('doAlignmentStep')
        else:
            self._insertFunctionStep('copyOpticalFlows')
        self._insertFunctionStep('findCorrelationMatrix')
        if (self.WarpAndEstimate.get()):
            self._insertFunctionStep('warpByFlow')
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def convertInputStep(self):
        # Write a metadata with the volumes
        xmipp3.convert.writeSetOfVolumes(self.inputVolumes.get(), self.imgsFn)

    def doAlignmentStep(self):
        imgFn = self.imgsFn
        StartingReference = self.StartingReference.get()
        ReferenceVolume = self.ReferenceVolume.get()

        if StartingReference == REFERENCE_STA:
            STAVolume = self.STAVolume.get().getFileName()
        else:
            STAVolume = ReferenceVolume
        # just in case the reference is in MRC format:
        path_vol0 = self._getExtraPath('reference.spi')
        params = '-i ' + STAVolume + ' -o ' + path_vol0 + ' --type vol'
        self.runJob('xmipp_image_convert', params)

        pyr_scale = self.pyr_scale.get()
        levels = self.levels.get()
        iterations = self.iterations.get()
        winsize = self.winsize.get()
        poly_n = self.poly_n.get()
        poly_sigma = self.poly_sigma.get()
        # TODO: the factor1 and 2 can be any value as long as we are using the subtomogram average (gray level values
        # are similar. It is not sure if we use an external reference what this should be! This could be normalized in
        #  future
        flags = self.flags.get()
        factor1 = self.factor1.get()
        factor2 = self.factor2.get()

        mdImgs = md.MetaData(imgFn)
        of_root = self._getExtraPath() + '/optical_flows/'

        # This is a spherical mask with maximum radius
        mask_size = int(self.getVolumeDimesion()//2)
        # Parallel processing (finding multiple optical flows at the same time)
        GPUids = np.array(getListFromRangeString(self.GPU_list.get()))
        gpu_ps = np.tile(GPUids, mdImgs.size())

        global segment
        def segment(objId):
            gpu_p = gpu_ps[objId-1]
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            # getting a copy converted to spider format to solve the problem with stacks or mrc files
            tmp = self._getTmpPath('tmp_' + str(objId) + '.spi')
            runProgram('xmipp_image_convert', '-i ' + imgPath + ' -o ' + tmp + ' --type vol')

            print('processing optical flow for volume ', objId)
            path_flowx = of_root + str(objId).zfill(6) + '_opflowx.spi'
            path_flowy = of_root + str(objId).zfill(6) + '_opflowy.spi'
            path_flowz = of_root + str(objId).zfill(6) + '_opflowz.spi'
            path_vol_i = tmp
            if (isfile(path_flowx)):
                return
            else:
                args = " %s %s %f %d %d %d %d %f %d %d %s %s %s %d" % (path_vol0, path_vol_i, pyr_scale, levels, winsize,
                                                                   iterations, poly_n, poly_sigma, factor1, factor2,
                                                                   path_flowx, path_flowy, path_flowz, gpu_p)
                script_path = continuousflex.__path__[0] + '/protocols/utilities/optflow_run.py'
                command = "python " + script_path + args
                command = Plugin.getContinuousFlexCmd(command)
                check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr,
                           env=None, cwd=None)

                arg_x = "-i %s  --mask circular -%d --substitute 0  -o %s" % (path_flowx, mask_size, path_flowx)
                arg_y = "-i %s  --mask circular -%d --substitute 0  -o %s" % (path_flowy, mask_size, path_flowy)
                arg_z = "-i %s  --mask circular -%d --substitute 0  -o %s" % (path_flowz, mask_size, path_flowz)
                runProgram('xmipp_transform_mask', arg_x)
                runProgram('xmipp_transform_mask', arg_y)
                runProgram('xmipp_transform_mask', arg_z)

        # Running the multiple processing:
        ps = [objId for objId in mdImgs]
        Parallel(n_jobs=self.N_GPU.get(), backend="multiprocessing")(delayed(segment)(p) for p in ps)

    def findCorrelationMatrix(self):
        imgFn = self.imgsFn
        mdImgs = md.MetaData(imgFn)
        N = 0
        for objId in mdImgs:
            N += 1
        metric_mat = np.zeros([N, N])
        for i in range(1, N + 1):
            print('finding the correlation matrix row ', i)
            flowi = self.read_optical_flow_by_number(i)
            for j in range(i, N + 1):
                print('        column', j)
                flowj = self.read_optical_flow_by_number(j)
                metric_mat[i - 1, j - 1] = metric_mat[j - 1, i - 1] =  self.metric_opflow_vols(flowi, flowj)

        correlation_matrix = self._getExtraPath('data.csv')
        np.savetxt(correlation_matrix, metric_mat, delimiter=',')

    def copyOpticalFlows(self):
        # In this case we get from the refinment protocol the optical flows and the reference
        N = self.refinementProt.get().NumOfIters.get()
        self.imgsFn = self.refinementProt.get()._getExtraPath('volumes_aligned_'+str(N+1)+'.xmd')
        createLink(self.refinementProt.get()._getExtraPath() + '/optical_flows_' + str(N),
                   self._getExtraPath() + '/optical_flows')
        createLink(self.refinementProt.get()._getExtraPath('reference'+str(N+1)+'.spi'),
                   self._getExtraPath('reference.spi'))

    def warpByFlow(self):
        makePath(self._getExtraPath() + '/estimated_volumes')
        estVol_root = self._getExtraPath() + '/estimated_volumes/'
        reference_fn = self._getExtraPath('reference.spi')
        # reference = open_volume(reference_fn)
        reference = ImageHandler().read(reference_fn).getData()

        stat_mat_fn = self._getExtraPath('cc_msd_mad.txt')
        # recount the number of volumes:
        imgFn = self.imgsFn
        mdImgs = md.MetaData(imgFn)

        N = 0
        for objId in mdImgs:
            N += 1

        # dumping the reference volume to use it in an outside script
        ref_dump = self._getTmpPath('ref_dump.pkl')
        joblib.dump(reference, ref_dump)

        # TODO: this loop can be parallelized, but it is not too computationally demanding
        for i in range(1, N + 1):
            print('Warping a copy of the reference volume by the optical flow ', i)
            flow_i = np.float32(self.read_optical_flow_by_number(i))
            # dumping the optical flow to use it in an outside script
            flow_dump = self._getTmpPath('flow_dump.pkl')
            joblib.dump(flow_i, flow_dump)
            warped_path_i = estVol_root + str(i).zfill(6) + '.spi'
            args = " %s %s %s" % (ref_dump, flow_dump, warped_path_i)
            script_path = continuousflex.__path__[0] + '/protocols/utilities/optflow_warp.py'
            command = "python " + script_path + args
            command = Plugin.getContinuousFlexCmd(command)
            check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr,
                       env=None, cwd=None)

        # Find a matrix of metrics (normalized cross correlation, mean square distance, mean absolute distance)
        stat_mat = np.zeros([N, 3])
        mdImgs = md.MetaData(imgFn)
        i = 0
        for objId in mdImgs:
            print('Finding the cross corrlation, mean square distance and mean absolute distance for volume ', i + 1)
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            # getting a copy converted to spider format to solve the problem with stacks or mrc files
            tmp = self._getTmpPath('tmp.spi')
            self.runJob('xmipp_image_convert', '-i ' + imgPath + ' -o ' + tmp + ' --type vol')
            # vol_i = open_volume(tmp)
            vol_i = ImageHandler().read(tmp).getData()

            warped_path_i = estVol_root + str(i + 1).zfill(6) + '.spi'
            warped_i = ImageHandler().read(warped_path_i).getData()
            # TODO: replace all the open_volume and save_volume by the ImageHandler() read and write
            # warped_i = open_volume(warped_path_i)
            stat_mat[i, 0] = self.ncc(warped_i,vol_i)
            stat_mat[i, 1] = self.vmsq(warped_i,vol_i)
            stat_mat[i, 2] = self.vmab(warped_i,vol_i)
            i += 1
        np.savetxt(stat_mat_fn, stat_mat)
        pass

    def createOutputStep(self):
        if (self.WarpAndEstimate.get()):
            # first making a metadata for the wrapped volumes:
            out_mdfn = self._getExtraPath('volumes_out.xmd')
            pattern = '"' + self._getExtraPath() + '/estimated_volumes/*.spi"'
            command = '-p ' + pattern + ' -o ' + out_mdfn
            self.runJob('xmipp_metadata_selfile_create',command)
            # now creating the output set of volumes:
            partSet = self._createSetOfVolumes('Warped')
            xmipp3.convert.readSetOfVolumes(out_mdfn, partSet)
            if (self.copy_opflows.get() == FIND_FLOWS):
                partSet.setSamplingRate(self.inputVolumes.get().getSamplingRate())
            else:
                partSet.setSamplingRate(self.refinementProt.get().RefinedAverage.getSamplingRate())
            self._defineOutputs(WarpedRefByFlows=partSet)
            pass
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def read_optical_flow(self, path_flowx, path_flowy, path_flowz):
        x = ImageHandler().read(path_flowx).getData()
        y = ImageHandler().read(path_flowy).getData()
        z = ImageHandler().read(path_flowz).getData()

        l = np.shape(x)
        # print(l)
        flow = np.zeros([3, l[0], l[1], l[2]])
        flow[0, :, :, :] = x
        flow[1, :, :, :] = y
        flow[2, :, :, :] = z
        return flow

    def read_optical_flow_by_number(self, num):
        op_path = self._getExtraPath() + '/optical_flows/'
        path_flowx = op_path + str(num).zfill(6) + '_opflowx.spi'
        path_flowy = op_path + str(num).zfill(6) + '_opflowy.spi'
        path_flowz = op_path + str(num).zfill(6) + '_opflowz.spi'
        flow = self.read_optical_flow(path_flowx, path_flowy, path_flowz)
        return flow

    def metric_opflow_vols(self, flow1, flow2):
        # print(np.shape(flow1))
        reshaped_flow1 = np.reshape(flow1, [3, np.shape(flow1)[1] * np.shape(flow1)[2] * np.shape(flow1)[3]])
        reshaped_flow2 = np.reshape(flow2, [3, np.shape(flow2)[1] * np.shape(flow2)[2] * np.shape(flow2)[3]])
        metric = np.sum(np.multiply(reshaped_flow1, reshaped_flow2))
        return metric

    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd',
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def normalize(self, v):
        """Normalize the data.
        @param v: input volume.
        @return: Normalized volume.
        """
        m = np.mean(v)
        v = v - m
        s = np.std(v)
        v = v / s
        return v

    def ncc(self, v1, v2):
        """Compute the Normalized Cross Correlation between the two volumes.
        @param v1: volume 1.
        @param v2: volume 2.
        @return: NCC.
        """
        vv1 = self.normalize(v1)
        vv2 = self.normalize(v2)
        score = np.sum(vv1 * vv2) / vv1.size
        return score

    def vmsq(self, v1, v2):
        """Compute the normalized mean square distance between the two volumes
        :param v1: volume1
        :param v2: volume2
        :return: score for mean square diff
        """
        from sklearn.metrics import mean_squared_error
        score = mean_squared_error(np.ndarray.flatten(v1), np.ndarray.flatten(v2)) / mean_squared_error(
            np.ndarray.flatten(v1),
            np.zeros([np.size(np.ndarray.flatten(v1))]))
        return score

    def vmab(self, v1, v2):
        """Compute the normalized mean absolute distance between the two volumes
        :param v1: volume1
        :param v2: volume2
        :return: score for mean square diff
        """
        from sklearn.metrics import mean_absolute_error
        score = mean_absolute_error(np.ndarray.flatten(v1), np.ndarray.flatten(v2)) / mean_absolute_error(
            np.ndarray.flatten(v1),
            np.zeros([np.size(np.ndarray.flatten(v1))]))
        return score

    def getVolumeDimesion(self):
        return self.inputVolumes.get().getDimensions()[0]

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _citations(self):
        return ['harastani2022tomoflow','harastani2022continuousflex']

    def _methods(self):
        pass


# ***************************************************************************
# * Authors:     Mohsen Kazemi (mkazemi@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
from pyworkflow.tests import BaseTest, setupTestProject, DataSet
from pwem.protocols import ProtImportVolumes
from continuousflex.protocols import FlexProtStructureMapping


class TestStructureMapping(BaseTest):
  @classmethod
  def setUpClass(cls):
    setupTestProject(cls)
    cls.dsRelion = DataSet.getDataSet('relion_tutorial')
    cls.protImport1 = cls.newProtocol(ProtImportVolumes,
                                      filesPath=cls.dsRelion.getFile('volumes/reference_rotated_masked.vol'),
                                      samplingRate=1.0)
    cls.launchProtocol(cls.protImport1)
    cls.protImport2 = cls.newProtocol(ProtImportVolumes,
                                      filesPath=cls.dsRelion.getFile('volumes/reference_masked.vol'),
                                      samplingRate=1.0)
    cls.launchProtocol(cls.protImport2)

  def testStructureMapping(self):
    """
        Test protocol structure mapping
    """
    protStrucMap = self.newProtocol(FlexProtStructureMapping,
                                    objLabel='structure mapping',
                                    numberOfMpi=1, numberOfThreads=4)
    protStrucMap.inputVolumes.append(self.protImport1.outputVolume)
    protStrucMap.inputVolumes.append(self.protImport2.outputVolume)
    protStrucMap.numberOfModes.set(20)
    protStrucMap.pseudoAtomTarget.set(2.0)
    protStrucMap.pseudoAtomRadius.set(0.75)
    protStrucMap.rcPercentage.set(90.0)
    protStrucMap.collectivityThreshold.set(0.2)
    self.launchProtocol(protStrucMap)


# **************************************************************************
# *
# * Authors:
# * Ilyes Hamitouche
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.protocol.params import PointerParam
import pyworkflow.protocol.params as params
from pwem.protocols import ProtAnalysis3D
from subprocess import check_call
import sys
import continuousflex
from pyworkflow import BETA
from continuousflex import Plugin

OPTION_NMA = 0
OPTION_ANGLES = 1
OPTION_SHFITS = 2
OPTION_ALL = 3

DEVICE_CUDA = 0
DEVICE_CPU = 1


class FlexProtDeepHEMNMATrain(ProtAnalysis3D):
    """ DeepHEMNMA protocol, a neural network that learns the rigid-body parameters and the normal mode
        amplitudes estimated by HEMNMA protocol.
    """
    _label = 'deep hemnma train'
    _devStatus = BETA

    #--------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('analyze_option', params.EnumParam, label='set the parameter to train on',
                      display=params.EnumParam.DISPLAY_COMBO,
                      choices=['Train on Normal Mode Amplitudes',
                               'Train on Angles',
                               'Train on Shifts',
                               'Train on All parameters',
                               ], default = OPTION_ALL,
                      help='select a set of parameter to train on')
        group = form.addGroup('Train on conformational variability', condition='analyze_option == %d or analyze_option == %d'% (OPTION_NMA, OPTION_ALL))
        group.addParam('inputNMA', PointerParam, pointerClass='FlexProtAlignmentNMA',
                      label="Previous HEMNMA run",
                      help='Select a previous run of the NMA image alignment.', allowsNull=True)
        group = form.addGroup('Train on rigid-body variability ', condition='analyze_option == %d or analyze_option == %d' %(OPTION_SHFITS, OPTION_ANGLES))
        group.addParam('inputParticles', PointerParam, pointerClass='SetOfParticles',
                      label="Preious run of rigid-body alignment",
                      help='Select a previous run of rigid-body alignment.', allowsNull=True)
        form.addParam('device_option', params.EnumParam, label='set the device for training',
                      display=params.EnumParam.DISPLAY_COMBO,
                      choices=['train on GPUs',
                               'tain on CPUs'], default = DEVICE_CUDA,
                      help='set a device to run the training on')
        form.addParam('learning_rate', params.FloatParam, label = 'Learning rate', default = 0.0001)
        form.addParam('epochs', params.IntParam, expertLevel=params.LEVEL_ADVANCED,label = 'Number of epochs', default = 400)
        form.addParam('batch_size', params.IntParam ,expertLevel=params.LEVEL_ADVANCED, label = 'Batch size', default = 2)
        form.addParallelSection(threads=0, mpi=0)

    
    #--------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        self._insertFunctionStep('performDeepHEMNMAStep')
        self._insertFunctionStep('createOutputStep')
        
    #--------------------------- STEPS functions --------------------------------------------

    def performDeepHEMNMAStep(self):
        epochs = self.epochs.get()
        batch_size = self.batch_size.get()
        lr = self.learning_rate.get()
        mode = self.analyze_option.get()
        device = self.device_option.get()
        imgsFn = self.inputNMA.get()._getExtraPath('images.xmd')

        params = " %s %s %d %d %f %d %d" % (imgsFn, self._getExtraPath(), epochs, batch_size, lr, mode, device)
        script_path = continuousflex.__path__[0]+'/protocols/utilities/deep_hemnma.py'
        command = "python " + script_path + params
        command = Plugin.getContinuousFlexCmd(command)
        check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr, env=None, cwd=None)


    def createOutputStep(self):
        pass

    #--------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary
    
    def _validate(self):
        errors = []
        return errors
    
    def _citations(self):
        return ['harastani2022continuousflex','hamitouche2022deephemnma']
    
    def _methods(self):
        return []
    
    #--------------------------- UTILS functions --------------------------------------------

    def getInputParticles(self):
        """ Get the output particles of the input NMA protocol. """
        return self.inputNMA.get().outputParticles

    def getParticlesMD(self):
        "Get the metadata files that contain the NMA displacement"
        return self.inputNMA.get()._getExtraPath('images.xmd')

    def getInputPdb(self):
        return self.inputNMA.get().getInputPdb()
    
    def getOutputMatrixFile(self):
        return self._getExtraPath('output_matrix.txt')
    
    def getDeformationFile(self):
        return self._getExtraPath('deformations.txt')

    def getProjectorFile(self):
        return self.mappingFile.get()




"""
This file reads and displays SPIDER 
'single' image using PIL (pillow) module 

"""

# import PIL module
from PIL import Image
import glob as glob
import numpy as np
import matplotlib.pyplot as plt
#import mrcfile
from tqdm import tqdm
import torch
from pwem.emlib.image import ImageHandler

"""
def spi2array(f_name) -> object:
    spi_image = Image.open(f_name, 'r')
    spi_array = np.array(spi_image, dtype='float32')
    spi_array = normalize(spi_array)
    return spi_array
"""
def spi2array(f_name) -> object:
    spi_array = ImageHandler().read(f_name).getData()
    return spi_array


# read SPIDER dataset from directory
def read_from_directory(dataset_path) -> object:
    spi_dataset = []
    for img_path in glob.glob(dataset_path + '/*.spi'):
        spi_dataset.append(spi2array(img_path))
    return spi_dataset


# read SPIDER dataset from list
def read_from_list(xmd_path, img_name_list):
    spi_dataset = []
    for img_path in img_name_list:
        spi_dataset.append(spi2array(str(xmd_path) + '/' + img_path))
    return spi_dataset


# show SPIDER image format
def imshow(spi_image):
    plt.imshow(spi_image)
    plt.show()
    return

def normalize(spi_array):
    _sdv = np.std(spi_array)
    _mean = np.mean(spi_array)
    _min = np.min(spi_array)
    _max = np.max(spi_array)
    spi_array = (spi_array)/max(_max, np.abs(_min))
    #spi_array = (spi_array - _mean) / _sdv
    #spi_array = (spi_array - _min) / (_max - _min)
    return spi_array
    
    
def torch_normalize(spi_array):
    _sdv = torch.std(spi_array)
    _mean = torch.mean(spi_array)
    _min = torch.min(spi_array)
    _max = torch.max(spi_array)
    spi_array = (spi_array - _min) / (_max - _min)
    return spi_array

"""
def mrc_stack_reader(path):
    mrc = mrcfile.mmap(path, mode='r+')
    return mrc
"""



# **************************************************************************
# *
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *             Slavica Jonic                (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import basename
import os
from pyworkflow.utils import getListFromRangeString
from pwem.protocols import ProtAnalysis3D
from xmipp3.convert import (writeSetOfVolumes, xmippToLocation, createItemMatrix,
                            setXmippAttributes, getImageLocation)
import pwem as em
import pwem.emlib.metadata as md
from xmipp3 import XmippMdRow
from pyworkflow.utils.path import copyFile, cleanPath
import pyworkflow.protocol.params as params
from pyworkflow.protocol.params import NumericRangeParam
from .convert import modeToRow, eulerAngles2matrix, matrix2eulerAngles
from pwem import Domain
import numpy as np
import multiprocessing

WEDGE_MASK_NONE = 0
WEDGE_MASK_THRE = 1


class FlexProtAlignmentNMAVol(ProtAnalysis3D):
    """ Protocol for rigid-body and elastic alignment for volumes using NMA. This protocol is the code module of HEMNMA-3D.
     It will take as input a set of normal modes calculated for an input atomic or pseudoatomic structure, and a set of volumes (subtomograms) to analyze.
     It fits the input structure using its modes (a subset of the modes need to be selected) into each one of the input volumes while simultaneously looking for rigid-body alignment, with
     compensation for missing wedge artefacts.
     The result of this protocol are rigid-body and elastic parameters for each input volume.
     Those results will be fed for a dimensionality reduction method (nma dimred vol) for further processing.
     """
    _label = 'nma alignment vol'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputModes', params.PointerParam, pointerClass='SetOfNormalModes',
                      label="Normal modes",
                      help='Set of modes computed by normal mode analysis.')
        form.addParam('modeList', NumericRangeParam, expertLevel=params.LEVEL_ADVANCED,
                      label="Modes selection",
                      help='Select the normal modes that will be used for volume analysis. \n'
                           'If you leave this field empty, all computed modes will be selected for image analysis.\n'
                           'You have several ways to specify the modes.\n'
                           '   Examples:\n'
                           ' "7,8-10" -> [7,8,9,10]\n'
                           ' "8, 10, 12" -> [8,10,12]\n'
                           ' "8 9, 10-12" -> [8,9,10,11,12])\n')
        form.addParam('inputVolumes', params.PointerParam,
                      pointerClass='SetOfVolumes,Volume',
                      label="Input volume(s)", important=True,
                      help='Select the set of volumes that will be analyzed using normal modes.')
        form.addParam('copyDeformations', params.PathParam,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='Precomputed results (for development)',
                      help='Enter a metadata file with precomputed elastic  \n'
                           'and rigid-body alignment parameters to perform \n'
                           'remaining steps using this file.')
        form.addSection(label='Missing-wedge Compensation')
        form.addParam('WedgeMode', params.EnumParam,
                      choices=['Do not compensate', 'Compensate'],
                      default=WEDGE_MASK_THRE,
                      label='Wedge mode', display=params.EnumParam.DISPLAY_COMBO,
                      help='Choose to compensate for the missing wedge if the data is subtomograms.'
                           ' However, if you correct the missing wedge in advance, then choose not to compensate.'
                           ' You can also choose not to compensate if your data is not subtomograms but EM-maps.'
                           ' The missing wedge is assumed to be in the Y-axis direction.')
        form.addParam('tiltLow', params.IntParam, default=-60,
                      condition='WedgeMode==%d' % WEDGE_MASK_THRE,
                      label='Lower tilt value',
                      help='The lower tilt angle used in obtaining the tilt series')
        form.addParam('tiltHigh', params.IntParam, default=60,
                      condition='WedgeMode==%d' % WEDGE_MASK_THRE,
                      label='Upper tilt value',
                      help='The upper tilt angle used in obtaining the tilt series')

        form.addSection(label='Combined elastic and rigid-body alignment')
        form.addParam('trustRegionScale', params.FloatParam, default=1.0,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='Elastic alignment trust region scale ',
                      help='For elastic alignment, this parameter scales the initial '
                           'value of the trust region radius of CONDOR optimization. '
                           'The default value of 1 works in majority of cases. \n'
                           'This value should not be changed except by expert users. '
                           'Larger values (e.g., between 1 and 2) can be tried '
                           'for larger expected amplitudes of conformational change.')
        form.addHidden('rhoStartBase', params.FloatParam, default=250.0,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='CONDOR optimiser parameter rhoStartBase',
                      help='rhoStartBase > 0  : (rhoStart = rhoStartBase*trustRegionScale) the lower the better,'
                           ' yet the slower')
        form.addHidden('rhoEndBase', params.FloatParam, default=50.0,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='CONDOR optimiser parameter rhoEndBase ',
                      help='rhoEndBase > 250  : (rhoEnd = rhoEndBase*trustRegionScale) no specific rule, '
                           'however it is better to keep it < 1000 if set very high we risk distortions')
        form.addHidden('niter', params.IntParam, default=10000,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='CONDOR optimiser parameter niter',
                      help='niter should be big enough to guarantee that the search converges to the '
                           'right set of nma deformation amplitudes')
        form.addParam('frm_freq', params.FloatParam, default=0.25,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='Maximum cross correlation frequency',
                      help='The normalized frequency should be between 0 and 0.5 \n'
                           'The larger it is, the bigger the search frequency is, the more time it demands. '
                           'Keeping it as default is recommended.')
        form.addParam('frm_maxshift', params.IntParam, default=10,
                      expertlevel=params.LEVEL_ADVANCED,
                      label='Maximum shift for rigid body alignment (in pixels)',
                      help='The maximum shift is a number between 1 and half the size of your volume. '
                           'It represents the maximum distance searched in x, y and z directions. Keep as default'
                           ' if your target is near the center in your subtomograms')
        form.addParallelSection(threads=0, mpi=multiprocessing.cpu_count()//2-1)

    # --------------------------- INSERT steps functions --------------------------------------------
    def getInputPdb(self):
        """ Return the Pdb object associated with the normal modes. """
        return self.inputModes.get().getPdb()

    def _insertAllSteps(self):
        atomsFn = self.getInputPdb().getFileName()
        # Define some outputs filenames
        self.imgsFn = self._getExtraPath('volumes.xmd')
        self.imgsFn_backup = self._getExtraPath('volumes_backup.xmd')
        self.modesFn = self._getExtraPath('modes.xmd')
        self.structureEM = self.inputModes.get().getPdb().getPseudoAtoms()
        if self.structureEM:
            self.atomsFn = self._getExtraPath(basename(atomsFn))
            copyFile(atomsFn, self.atomsFn)
        else:
            pdb_name = os.path.dirname(self.inputModes.get().getFileName()) + '/atoms.pdb'
            self.atomsFn = self._getExtraPath(basename(pdb_name))
            copyFile(pdb_name, self.atomsFn)

        self._insertFunctionStep('convertInputStep', atomsFn)

        if self.copyDeformations.empty():
            self._insertFunctionStep("performNmaStep", self.atomsFn, self.modesFn)
        else:  # SERVES FOR DEBUGGING AND COMPUTING ON CLUSTERS
            self._insertFunctionStep('copyDeformationsStep', self.copyDeformations.get())

        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def convertInputStep(self, atomsFn):
        # Write the modes metadata taking into account the selection
        self.writeModesMetaData()
        # Write a metadata with the normal modes information
        # to launch the nma alignment programs
        writeSetOfVolumes(self.inputVolumes.get(), self.imgsFn)
        writeSetOfVolumes(self.inputVolumes.get(), self.imgsFn_backup)

    def writeModesMetaData(self):
        """ Iterate over the input SetOfNormalModes and write
        the proper Xmipp metadata.
        Take into account a possible selection of modes
        """

        if self.modeList.empty():
            modeSelection = []
        else:
            modeSelection = getListFromRangeString(self.modeList.get())

        mdModes = md.MetaData()

        inputModes = self.inputModes.get()
        for mode in inputModes:
            # If there is a mode selection, only
            # take into account those selected
            if not modeSelection or mode.getObjId() in modeSelection:
                row = XmippMdRow()
                modeToRow(mode, row)
                row.writeToMd(mdModes, mdModes.addObject())
        mdModes.write(self.modesFn)

    def copyDeformationsStep(self, deformationMd):
        copyFile(deformationMd, self.imgsFn)
        # We update the volume paths based on volume names (if computed on another computer or imported from another
        # project), and we need to set the item_id for each volume
        inputSet = self.inputVolumes.get()
        mdImgs = md.MetaData(self.imgsFn)
        for objId in mdImgs:
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            index, fn = xmippToLocation(imgPath)
            if(index): # case the input is a stack
                # Conside the index is the id in the input set
                particle = inputSet[index]
            else: # input is not a stack
                # convert the inputSet to metadata:
                mdtemp = md.MetaData(self.imgsFn_backup)
                # Loop and find the index based on the basename:
                bn_retrieved = basename(imgPath)
                for searched_index in mdtemp:
                    imgPath_temp = mdtemp.getValue(md.MDL_IMAGE,searched_index)
                    bn_searched = basename(imgPath_temp)
                    if bn_searched == bn_retrieved:
                        index = searched_index
                        particle = inputSet[index]
                        break
            mdImgs.setValue(md.MDL_IMAGE, getImageLocation(particle), objId)
            mdImgs.setValue(md.MDL_ITEM_ID, int(particle.getObjId()), objId)
        mdImgs.sort(md.MDL_ITEM_ID)
        mdImgs.write(self.imgsFn)

        # if the volumes were aligned with angle_y=90 degrees, then rotate by 90 and inverse, then set angle y to 0
        mdImgs = md.MetaData(self.imgsFn)

        flag = None
        try:
            flag = mdImgs.getValue(md.MDL_ANGLE_Y, 1)
        except:
            pass

        if flag == 90:
            mdImgs = md.MetaData(self.imgsFn)
            for objId in mdImgs:
                rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
                tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
                psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)
                x = mdImgs.getValue(md.MDL_SHIFT_X, objId)
                y = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
                z = mdImgs.getValue(md.MDL_SHIFT_Z, objId)
                T = eulerAngles2matrix(rot, tilt, psi, x, y, z)
                # Rotate 90 degrees (compensation for missing wedge)
                T0 = eulerAngles2matrix(0, 90, 0, 0, 0, 0)
                T = np.linalg.inv(np.matmul(T, T0))
                rot, tilt, psi, x, y, z = matrix2eulerAngles(T)
                mdImgs.setValue(md.MDL_ANGLE_ROT, rot, objId)
                mdImgs.setValue(md.MDL_ANGLE_TILT, tilt, objId)
                mdImgs.setValue(md.MDL_ANGLE_PSI, psi, objId)
                mdImgs.setValue(md.MDL_SHIFT_X, x, objId)
                mdImgs.setValue(md.MDL_SHIFT_Y, y, objId)
                mdImgs.setValue(md.MDL_SHIFT_Z, z, objId)
                mdImgs.setValue(md.MDL_ANGLE_Y, 0.0, objId)
            mdImgs.write(self.imgsFn)



    def performNmaStep(self, atomsFn, modesFn):
        sampling = self.inputVolumes.get().getSamplingRate()
        trustRegionScale = self.trustRegionScale.get()
        odir = self._getTmpPath()
        imgFn = self.imgsFn
        frm_freq = self.frm_freq.get()
        frm_maxshift = self.frm_maxshift.get()
        rhoStartBase = self.rhoStartBase.get()
        rhoEndBase = self.rhoEndBase.get()
        niter = self.niter.get()
        rhoStartBase = 250.0
        rhoEndBase = 50.0
        niter = 10000

        args = "-i %(imgFn)s --pdb %(atomsFn)s --modes %(modesFn)s --sampling_rate %(sampling)f "
        args += "--odir %(odir)s --centerPDB "
        args += "--trustradius_scale %(trustRegionScale)d --resume "

        if self.getInputPdb().getPseudoAtoms():
            args += "--fixed_Gaussian "

        args += "--alignVolumes %(frm_freq)f %(frm_maxshift)d "

        args += "--condor_params %(rhoStartBase)f %(rhoEndBase)f %(niter)d "

        if self.WedgeMode == WEDGE_MASK_THRE:
            tilt0 = self.tiltLow.get()
            tiltF = self.tiltHigh.get()
            args += "--tilt_values %(tilt0)d %(tiltF)d "

        self.runJob("xmipp_nma_alignment_vol", args % locals(),
                    env=Domain.importFromPlugin('xmipp3').Plugin.getEnviron())

        cleanPath(self._getPath('nmaTodo.xmd'))

        inputSet = self.inputVolumes.get()
        mdImgs = md.MetaData(self.imgsFn)

        for objId in mdImgs:
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            index, fn = xmippToLocation(imgPath)
            if(index): # case the input is a stack
                # Conside the index is the id in the input set
                particle = inputSet[index]
            else: # input is not a stack
                # convert the inputSet to metadata:
                mdtemp = md.MetaData(self.imgsFn_backup)
                # Loop and find the index based on the basename:
                bn_retrieved = basename(imgPath)
                for searched_index in mdtemp:
                    imgPath_temp = mdtemp.getValue(md.MDL_IMAGE,searched_index)
                    bn_searched = basename(imgPath_temp)
                    if bn_searched == bn_retrieved:
                        index = searched_index
                        particle = inputSet[index]
                        break
            mdImgs.setValue(md.MDL_IMAGE, getImageLocation(particle), objId)
            mdImgs.setValue(md.MDL_ITEM_ID, int(particle.getObjId()), objId)
        mdImgs.sort(md.MDL_ITEM_ID)
        mdImgs.write(self.imgsFn)

        # if WedgeMode was Mask, then update the metadata angles and shifts to be in the same convention of the
        # ground truth (rotate 90 degrees) then set angle y to 0
        if self.WedgeMode == WEDGE_MASK_THRE:
            mdImgs = md.MetaData(self.imgsFn)
            for objId in mdImgs:
                rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
                tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
                psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)
                x = mdImgs.getValue(md.MDL_SHIFT_X, objId)
                y = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
                z = mdImgs.getValue(md.MDL_SHIFT_Z, objId)
                T = eulerAngles2matrix(rot, tilt, psi, x, y, z)
                # Rotate 90 degrees (compensation for missing wedge)
                T0 = eulerAngles2matrix(0, 90, 0, 0, 0, 0)
                T = np.linalg.inv(np.matmul(T, T0))
                rot, tilt, psi, x, y, z = matrix2eulerAngles(T)
                mdImgs.setValue(md.MDL_ANGLE_ROT, rot, objId)
                mdImgs.setValue(md.MDL_ANGLE_TILT, tilt, objId)
                mdImgs.setValue(md.MDL_ANGLE_PSI, psi, objId)
                mdImgs.setValue(md.MDL_SHIFT_X, x, objId)
                mdImgs.setValue(md.MDL_SHIFT_Y, y, objId)
                mdImgs.setValue(md.MDL_SHIFT_Z, z, objId)
                mdImgs.setValue(md.MDL_ANGLE_Y, 0.0, objId)
            mdImgs.write(self.imgsFn)

        cleanPath(self._getExtraPath('copy.xmd'))

    def createOutputStep(self):
        inputSet = self.inputVolumes.get()
        # partSet = self._createSetOfParticles()
        partSet = self._createSetOfVolumes()
        pdbPointer = self.inputModes.get()._pdbPointer

        partSet.copyInfo(inputSet)
        partSet.setAlignmentProj()
        partSet.copyItems(inputSet,
                          updateItemCallback=self._updateParticle,
                          itemDataIterator=md.iterRows(self.imgsFn, sortByLabel=md.MDL_ITEM_ID))

        self._defineOutputs(outputParticles=partSet)
        self._defineSourceRelation(pdbPointer, partSet)
        self._defineTransformRelation(self.inputVolumes, partSet)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2021hemnma','harastani2022continuousflex']

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def _updateParticle(self, item, row):
        setXmippAttributes(item, row, md.MDL_ANGLE_ROT, md.MDL_ANGLE_TILT, md.MDL_ANGLE_PSI, md.MDL_SHIFT_X,
                           md.MDL_SHIFT_Y, md.MDL_SHIFT_Z, md.MDL_FLIP, md.MDL_NMA, md.MDL_COST, md.MDL_MAXCC)
        createItemMatrix(item, row, align=em.ALIGN_PROJ)



# **************************************************************************
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pwem.protocols import ProtAnalysis3D
import xmipp3.convert
import pwem.emlib.metadata as md
import pyworkflow.protocol.params as params
from pyworkflow.utils.path import makePath, copyFile
from os.path import basename, isfile
from pyworkflow.utils import replaceBaseExt
from continuousflex.protocols.utilities.bm4d import bm4d
from pwem.utils import runProgram


REFERENCE_EXT = 0
REFERENCE_STA = 1

METHOD_BM4D = 0
METHOD_LOWPASS = 1

NOISE_GAUSS = 0
NOISE_RICE = 1

PROFILE_LC = 0
PROFILE_NP = 1
PROFILE_MP = 2


class FlexProtVolumeDenoise(ProtAnalysis3D):
    """ Protocol for subtomogram missingwedge filling. """
    _label = 'volume denoise'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputVolumes', params.PointerParam,
                      pointerClass='SetOfVolumes,Volume',
                      label="Input volume(s)", important=True,
                      help='Select volumes')
        form.addSection('Method')
        form.addParam('Method', params.EnumParam,
                      choices=['BM4D', 'Fourier lowpass filter'],
                      default=METHOD_BM4D,
                      label='Denoising Method', display=params.EnumParam.DISPLAY_COMBO,
                      help='Choose a method: BM4D or Fourier lowpass filter')
        group = form.addGroup('BM4D parameters', condition='Method==%d' % METHOD_BM4D)
        group.addParam('noiseType', params.EnumParam,
                      choices=['Gaussian', 'Rician'],
                      default=NOISE_GAUSS,
                      label='Noise distribution', display=params.EnumParam.DISPLAY_COMBO,
                      help='Noise distribution (either Gaussian or Rician)')
        group.addParam('sigma_choice', params.EnumParam,
                      choices=['Automatically estimate sigma', 'Set a value for sigma (recommended)'],
                      default=1,
                      label='Sigma choice', display=params.EnumParam.DISPLAY_COMBO,
                      help='Sigma is the standard deviation of data noise')
        group.addParam('sigma', params.FloatParam, default=0.2, allowsNull=True,
                      condition='sigma_choice==%d' % 1,
                      label='Sigma',
                      help='estimated standard deviation of data noise '
                           'defines the strength of the processing (high value gives smooth images)')
        group.addParam('profile', params.EnumParam,
                      choices=['low complexity profile', 'normal profile', 'modified profile (recommended)'],
                      default=PROFILE_MP,
                      label='Noise profile', display=params.EnumParam.DISPLAY_COMBO,
                      help='lc --> low complexity profile, '
                           ' np --> normal profile,'
                           ' mp --> modified profile')
        group.addParam('do_wiener', params.BooleanParam, allowsNull=True,
                      default=False,
                      label='Do wiener?',
                      help='Perform collaborative Wiener filtering')

        # Normalized frequencies ("digital frequencies")
        line = form.addLine('Frequency (normalized)',
                            condition='Method==%d' % METHOD_LOWPASS,
                            help='The cufoff frequency and raised coside width of the low pass filter.'
                                 ' For details: see "xmipp_transform_filter --fourier low_pass"')
        line.addHidden('lowFreqDig', params.DigFreqParam, default=0.00, allowsNull=True,
                        label='Lowest')
        line.addParam('highFreqDig', params.DigFreqParam, default=0.25, allowsNull=True,
                      label='Cutoff frequency (0 -> 0.5)')
        line.addParam('freqDecayDig', params.FloatParam, default=0.02, allowsNull=True,
                      label='Raised cosine width')


    # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        # Define some outputs filenames
        self.imgsFn = self._getExtraPath('volumes.xmd')
        makePath(self._getExtraPath() + '/filtered')
        self._insertFunctionStep('convertInputStep')
        if(self.Method.get()==METHOD_BM4D):
            self._insertFunctionStep('denoise_b4md')
        else:
            self._insertFunctionStep('filter_lowpass')
        self._insertFunctionStep('createOutputStep')
        pass

    # --------------------------- STEPS functions --------------------------------------------
    def convertInputStep(self):
        # Write a metadata with the volumes
        try:
            xmipp3.convert.writeSetOfVolumes(self.inputVolumes.get(), self.imgsFn)
        except:
            mdF = md.MetaData()
            mdF.setValue(md.MDL_IMAGE, self.inputVolumes.get().getFileName(), mdF.addObject())
            mdF.write(self.imgsFn)
            pass

    def denoise_b4md(self):
        distribution = ''
        if self.noiseType.get() == NOISE_GAUSS:
            distribution = 'Gauss'
        else:
            distribution = 'Rice'
        sigma = self.sigma.get()
        profile = ''
        if self.profile.get()==PROFILE_LC:
            profile = 'lc'
        elif self.profile.get()==PROFILE_NP:
            profile = 'np'
        elif self.profile.get()==PROFILE_MP:
            profile = 'mp'
        else:
            exit()
        do_weiner = 0
        if self.do_wiener.get():
            do_weiner = 1


        tempdir = self._getTmpPath()
        imgFn = self.imgsFn
        # looping on all images and performing mwr
        mdImgs = md.MetaData(imgFn)
        for objId in mdImgs:
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            index, fname = xmipp3.convert.xmippToLocation(imgPath)
            new_imgPath = self._getExtraPath() + '/filtered/'
            if index:  # case of stack
                new_imgPath += str(index).zfill(6) + '.spi'
            else:
                new_imgPath += basename(replaceBaseExt(basename(imgPath), 'spi'))
            # Get a copy of the volume converted to spider format
            temp_path = self._getTmpPath('temp.spi')
            # params = '-i ' + imgPath + ' -o ' + new_imgPath + ' --type vol'
            params = '-i ' + imgPath + ' -o ' + temp_path + ' --type vol'
            runProgram('xmipp_image_convert', params)

            # perform the mwr:
            # in case the file exists (continuing or injecting)
            if (isfile(new_imgPath)):
                continue
            else:
                bm4d(temp_path, new_imgPath, distribution, sigma, profile, do_weiner)
            # update the name in the metadata file
            mdImgs.setValue(md.MDL_IMAGE, new_imgPath, objId)
        mdImgs.write(self.imgsFn)


    def filter_lowpass(self):
        cutoff = self.highFreqDig.get()
        raisedw = self.freqDecayDig.get()

        imgFn = self.imgsFn
        # looping on all images and performing mwr
        mdImgs = md.MetaData(imgFn)
        for objId in mdImgs:
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            index, fname = xmipp3.convert.xmippToLocation(imgPath)
            new_imgPath = self._getExtraPath() + '/filtered/'
            if index:  # case of stack
                new_imgPath += str(index).zfill(6) + '.spi'
            else:
                new_imgPath += basename(replaceBaseExt(basename(imgPath), 'spi'))
            # Get a copy of the volume converted to spider format
            temp_path = self._getTmpPath('temp.spi')
            # params = '-i ' + imgPath + ' -o ' + new_imgPath + ' --type vol'
            params = '-i ' + imgPath + ' -o ' + temp_path + ' --type vol'
            runProgram('xmipp_image_convert', params)

            # perform the mwr:
            # in case the file exists (continuing or injecting)
            if (isfile(new_imgPath)):
                continue
            else:
                params = " -i " + temp_path + " -o " + new_imgPath
                params += " --fourier low_pass " + str(cutoff) + ' ' + str(raisedw)
                runProgram('xmipp_transform_filter', params)
            # update the name in the metadata file
            mdImgs.setValue(md.MDL_IMAGE, new_imgPath, objId)
        mdImgs.write(self.imgsFn)

    def createOutputStep(self):
        partSet = self._createSetOfVolumes('filtered')
        xmipp3.convert.readSetOfVolumes(self._getExtraPath('volumes.xmd'), partSet)
        partSet.setSamplingRate(self.inputVolumes.get().getSamplingRate())
        self._defineOutputs(filteredVolumes=partSet)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd',
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()



# **************************************************************************
# * Authors:     Rémi Vuillemot (remi.vuillemot@upmc.fr)
# * IMPMC, Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pwem.protocols import ProtImportPdb, ProtImportVolumes#, ProtImportParticles, ProtImportVolumes
from pwem.tests.workflows import TestWorkflow
from pyworkflow.tests import setupTestProject, DataSet
from continuousflex.protocols.protocol_generate_topology import ProtGenerateTopology
from continuousflex.protocols import FlexProtNMA, NMA_CUTOFF_ABS, FlexProtSynthesizeImages
from continuousflex.viewers.viewer_genesis import *
from continuousflex.protocols.utilities.pdb_handler import ContinuousFlexPDBHandler


class testGENESIS(TestWorkflow):
    """ Test Class for GENESIS. """
    @classmethod
    def setUpClass(cls):
        # Create a new project
        setupTestProject(cls)
        cls.ds = DataSet.getDataSet('nma_V2.0')
        # Import Target EM map
        protImportVol = cls.newProtocol(ProtImportVolumes, importFrom=ProtImportVolumes.IMPORT_FROM_FILES,
                                       filesPath=cls.ds.getFile('1ake_vol'),  samplingRate=2.0)
        protImportVol.setObjLabel('Target EM volume (1AKE)')
        cls.launchProtocol(protImportVol)

        cls.protImportVol = protImportVol


    def test1_EmfitVolumeCHARMM(self):
        # Import PDB to fit
        protPdb4ake = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                         pdbFile=self.ds.getFile('4ake_aa_pdb'))
        protPdb4ake.setObjLabel('Input PDB (4AKE All-Atom)')
        self.launchProtocol(protPdb4ake)

        # Energy min
        protGenTopo = self.newProtocol(ProtGenerateTopology,
            inputPDB = protPdb4ake.outputPdb,
            forcefield = FORCEFIELD_CHARMM,
            inputPRM = self.ds.getFile('charmm_prm'),
            inputRTF = self.ds.getFile('charmm_top'),
            inputPSF=self.ds.getFile('4ake_aa_psf'))
        self.launchProtocol(protGenTopo)


        # Energy min
        protGenesisMin = self.newProtocol(ProtGenesis,
            inputType = INPUT_TOPOLOGY,
            topoProt = protGenTopo,

            simulationType = SIMULATION_MIN,
            time_step = 0.002,
            n_steps = 100,
            eneout_period = 10,
            crdout_period = 10,
            nbupdate_period = 10,

            implicitSolvent = IMPLICIT_SOLVENT_GBSA,
            electrostatics = ELECTROSTATICS_CUTOFF,
            switch_dist = 10.0,
            cutoff_dist = 12.0,
            pairlist_dist = 15.0,

            numberOfThreads = NUMBER_OF_CPU,
            numberOfMpi = 1,

       )

        protGenesisMin.setObjLabel('Energy Minimization CHARMM')
        # Launch minimisation
        self.launchProtocol(protGenesisMin)

        # Get GENESIS log file
        output_prefix = protGenesisMin.getOutputPrefix()
        log_file = output_prefix+".log"

        # Get the potential energy from the log file
        potential_ene = readLogFile(log_file)["POTENTIAL_ENE"]

        # Assert that the potential energy is decreasing
        print("\n\n//////////////////////////////////////////////")
        print(protGenesisMin.getObjLabel())
        print("Initial potential energy : %.2f kcal/mol"%potential_ene[0])
        print("Final potential energy : %.2f kcal/mol"%potential_ene[-1])
        print("//////////////////////////////////////////////\n\n")

        assert(potential_ene[0] > potential_ene[-1])


        # Launch NMA for energy min PDB
        protNMA = self.newProtocol(FlexProtNMA,
                                    cutoffMode=NMA_CUTOFF_ABS)
        protNMA.inputStructure.set(protGenesisMin.outputPDB)
        protNMA.setObjLabel('NMA')
        self.launchProtocol(protNMA)

        protGenesisFitNMMD = self.newProtocol(ProtGenesis,
          inputType=INPUT_RESTART,
          restartProt = protGenesisMin,

          simulationType=SIMULATION_NMMD,
          time_step=0.002,
          n_steps=100, # 3000
          eneout_period=100,
          crdout_period=100,
          nbupdate_period=10,
          nm_number=6,
          nm_mass=1.0,
          inputModes=protNMA.outputModes,

          implicitSolvent=IMPLICIT_SOLVENT_GBSA,
          electrostatics=ELECTROSTATICS_CUTOFF,
          switch_dist=10.0,
          cutoff_dist=12.0,
          pairlist_dist=15.0,

          ensemble=ENSEMBLE_NVT,
          tpcontrol=TPCONTROL_LANGEVIN,
          temperature=300.0,

          boundary=BOUNDARY_NOBC,
          EMfitChoice=EMFIT_VOLUMES,
          constantK=10000,
          emfit_sigma=2.0,
          emfit_tolerance=0.1,
          inputVolume=self.protImportVol.outputVolume,
          voxel_size=2.0,
          centerOrigin=True,

          numberOfThreads=NUMBER_OF_CPU,
          numberOfMpi=1,
          )
        protGenesisFitNMMD.setObjLabel('NMMD Flexible Fitting CHARMM')

        # Launch Fitting
        self.launchProtocol(protGenesisFitNMMD)

        # Get GENESIS log file
        log_file = protGenesisFitNMMD.getOutputPrefix()+".log"

        # Get the CC from the log file
        cc = readLogFile(log_file)["RESTR_CVS001"]

        # Get the RMSD
        inp = ContinuousFlexPDBHandler(protGenesisFitNMMD.getInputPDBprefix() + ".pdb")
        ref = ContinuousFlexPDBHandler(self.ds.getFile('1ake_pdb'))
        out = ContinuousFlexPDBHandler(protGenesisFitNMMD.getOutputPrefix()+".pdb")
        matchingAtoms = inp.matchPDBatoms(reference_pdb=ref)
        rmsd_inp = inp.getRMSD(reference_pdb=ref,idx_matching_atoms=matchingAtoms,align=True)
        rmsd_out = out.getRMSD(reference_pdb=ref,idx_matching_atoms=matchingAtoms,align=True)

        # Assert that the CC is increasing and  the RMSD is decreasing
        print("\n\n//////////////////////////////////////////////")
        print(protGenesisFitNMMD.getObjLabel())
        print("Initial CC : %.2f"%cc[0])
        print("Final CC : %.2f"%cc[-1])
        print("Initial rmsd : %.2f Ang"%rmsd_inp)
        print("Final rmsd : %.2f Ang"%rmsd_out)
        print("//////////////////////////////////////////////\n\n")

        assert(cc[0] < cc[-1])
        assert(rmsd_inp >rmsd_out)
        # assert(rmsd[-1] < 3.0)

    def test2_EmfitVolumeCAGO(self):
        # Import PDB to fit
        protPdb4ake = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                         pdbFile=self.ds.getFile('4ake_ca_pdb'))
        protPdb4ake.setObjLabel('Input PDB (4AKE C-Alpha only)')
        self.launchProtocol(protPdb4ake)


        protGenesisMin = self.newProtocol(ProtGenesis,
            inputPDB = protPdb4ake.outputPdb,
            forcefield = FORCEFIELD_CAGO,
            inputType = INPUT_NEW_SIM,
            inputTOP = self.ds.getFile('4ake_ca_top'),

            simulationType = SIMULATION_MIN,
            time_step = 0.001,
            n_steps = 100,
            eneout_period = 10,
            crdout_period = 10,
            nbupdate_period = 10,

            implicitSolvent = IMPLICIT_SOLVENT_NONE,
            electrostatics = ELECTROSTATICS_CUTOFF,
            switch_dist = 10.0,
            cutoff_dist = 12.0,
            pairlist_dist = 15.0,

          numberOfThreads=NUMBER_OF_CPU,
          numberOfMpi=1,
       )
        protGenesisMin.setObjLabel('Energy Minimization CAGO')
        # Launch minimisation
        self.launchProtocol(protGenesisMin)

        # Launch NMA for energy min PDB
        protNMA = self.newProtocol(FlexProtNMA,
                                    cutoffMode=NMA_CUTOFF_ABS)
        protNMA.inputStructure.set(protGenesisMin.outputPDB)
        protNMA.setObjLabel('NMA')
        self.launchProtocol(protNMA)

        protGenesisFitMD = self.newProtocol(ProtGenesis,

                                            inputType=INPUT_RESTART,
                                            restartProt=protGenesisMin,

                                              simulationType=SIMULATION_MD,
                                              time_step=0.0005,
                                              n_steps=1000,
                                              eneout_period=100,
                                              crdout_period=100,
                                              nbupdate_period=10,

                                              implicitSolvent=IMPLICIT_SOLVENT_NONE,
                                              electrostatics=ELECTROSTATICS_CUTOFF,
                                              switch_dist=10.0,
                                              cutoff_dist=12.0,
                                              pairlist_dist=15.0,

                                              ensemble=ENSEMBLE_NVT,
                                              tpcontrol=TPCONTROL_LANGEVIN,
                                              temperature=50.0,

                                              boundary=BOUNDARY_NOBC,
                                              EMfitChoice=EMFIT_VOLUMES,
                                              constantK="500",
                                              emfit_sigma=2.0,
                                              emfit_tolerance=0.1,
                                              inputVolume=self.protImportVol.outputVolume,
                                              voxel_size=2.0,
                                              centerOrigin=True,

                                              numberOfThreads=NUMBER_OF_CPU,
                                              numberOfMpi=1,
                                              )
        protGenesisFitMD.setObjLabel('MD Flexible Fitting CAGO')

        # Launch Fitting
        self.launchProtocol(protGenesisFitMD)

        # Get GENESIS log file
        log_file = protGenesisFitMD.getOutputPrefix()+".log"

        # Get the CC from the log file
        cc = readLogFile(log_file)["RESTR_CVS001"]

        # Get the RMSD
        inp = ContinuousFlexPDBHandler(protGenesisFitMD.getInputPDBprefix() + ".pdb")
        ref = ContinuousFlexPDBHandler(self.ds.getFile('1ake_pdb'))
        out = ContinuousFlexPDBHandler(protGenesisFitMD.getOutputPrefix()+".pdb")
        matchingAtoms = inp.matchPDBatoms(reference_pdb=ref)
        rmsd_inp = inp.getRMSD(reference_pdb=ref,idx_matching_atoms=matchingAtoms,align=True)
        rmsd_out = out.getRMSD(reference_pdb=ref,idx_matching_atoms=matchingAtoms,align=True)

        # Assert that the CC is increasing and  the RMSD is decreasing
        print("\n\n//////////////////////////////////////////////")
        print(protGenesisFitMD.getObjLabel())
        print("Initial CC : %.2f"%cc[0])
        print("Final CC : %.2f"%cc[-1])
        print("Initial rmsd : %.2f Ang"%rmsd_inp)
        print("Final rmsd : %.2f Ang"%rmsd_out)
        print("//////////////////////////////////////////////\n\n")
        assert (cc[0] < cc[-1])
        assert (rmsd_inp > rmsd_out)


        # Need at least 4 cores
        if NUMBER_OF_CPU >= 4:
            protGenesisFitREUS = self.newProtocol(ProtGenesis,

                                                  inputType=INPUT_RESTART,
                                                  restartProt=protGenesisMin,

                                                  simulationType=SIMULATION_RENMMD,
                                                  time_step=0.0005,
                                                  n_steps=1000,
                                                  eneout_period=100,
                                                  crdout_period=100,
                                                  nbupdate_period=10,
                                                  nm_number=6,
                                                  nm_mass=1.0,
                                                  inputModes=protNMA.outputModes,
                                                  exchange_period=100, # 100
                                                  nreplica = 4,

                                              implicitSolvent=IMPLICIT_SOLVENT_NONE,
                                              electrostatics=ELECTROSTATICS_CUTOFF,
                                              switch_dist=10.0,
                                              cutoff_dist=12.0,
                                              pairlist_dist=15.0,

                                              ensemble=ENSEMBLE_NVT,
                                              tpcontrol=TPCONTROL_LANGEVIN,
                                              temperature=50.0,

                                              boundary=BOUNDARY_NOBC,
                                              EMfitChoice=EMFIT_VOLUMES,
                                              constantK="500-1500",
                                              emfit_sigma=2.0,
                                              emfit_tolerance=0.1,
                                              inputVolume=self.protImportVol.outputVolume,
                                              voxel_size=2.0,
                                              centerOrigin=True,

                                                  numberOfThreads=1,
                                                  numberOfMpi=NUMBER_OF_CPU,
                                              )
            protGenesisFitREUS.setObjLabel('NMMD + REUS Flexible Fitting CAGO')

            # Launch Fitting
            self.launchProtocol(protGenesisFitREUS)

            # Get GENESIS log file
            outPref = protGenesisFitREUS.getOutputPrefixAll()
            log_file1 = outPref[0] + ".log"
            log_file2 = outPref[1] + ".log"

            # Get the CC from the log file
            cc1 = readLogFile(log_file1)["RESTR_CVS001"]
            cc2 = readLogFile(log_file2)["RESTR_CVS001"]

            # Get the RMSD
            ref = ContinuousFlexPDBHandler(self.ds.getFile('1ake_pdb'))
            inp = ContinuousFlexPDBHandler(protGenesisFitREUS.getInputPDBprefix() + ".pdb")
            out1 = ContinuousFlexPDBHandler(outPref[0] + ".pdb")
            out2 = ContinuousFlexPDBHandler(outPref[1] + ".pdb")
            matchingAtoms = inp.matchPDBatoms(reference_pdb=ref)
            rmsd_inp = inp.getRMSD(reference_pdb=ref, idx_matching_atoms=matchingAtoms, align=True)
            rmsd_out2 = out2.getRMSD(reference_pdb=ref, idx_matching_atoms=matchingAtoms, align=True)
            rmsd_out1 = out1.getRMSD(reference_pdb=ref, idx_matching_atoms=matchingAtoms, align=True)

            # Assert that the CCs are increasing
            print("\n\n//////////////////////////////////////////////")
            print(protGenesisFitREUS.getObjLabel())
            print("Initial CC : [%.2f , %.2f]" % (cc1[0],cc2[0]))
            print("Final CC :[%.2f , %.2f]" % (cc1[-1],cc2[-1]))
            print("Initial rmsd : [%.2f , %.2f] Ang" % (rmsd_inp,rmsd_inp))
            print("Final rmsd : [%.2f , %.2f] Ang" % (rmsd_out1,rmsd_out2))
            print("//////////////////////////////////////////////\n\n")

            assert (cc1[0] < cc1[-1])
            assert (cc2[0] < cc2[-1])
            assert (rmsd_inp> rmsd_out1)
            # assert (rmsd1[-1] < 3.0)
            assert (rmsd_inp > rmsd_out2)
            # assert (rmsd2[-1] < 3.0)

##################################################################################################
#
#                                  EMFIT IMAGES
#
##################################################################################################

            protPdb1ake = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                           pdbFile=self.ds.getFile('1ake_pdb'))
            protPdb1ake.setObjLabel('Target PDB (1AKE)')
            self.launchProtocol(protPdb1ake)
            protNMA_1ake = self.newProtocol(FlexProtNMA,
                                       cutoffMode=NMA_CUTOFF_ABS)
            protNMA_1ake.inputStructure.set(protPdb1ake.outputPdb)
            protNMA_1ake.setObjLabel('NMA 1ake')
            self.launchProtocol(protNMA_1ake)

            target_images= self.newProtocol(FlexProtSynthesizeImages,
                             inputModes=protNMA_1ake.outputModes,
                             numberOfVolumes=10,
                             samplingRate=2.0,
                             volumeSize=64)
            target_images.setObjLabel('Target particles (1ake)')
            self.launchProtocol(target_images)

            protGenesisFitNMMDImg = self.newProtocol(ProtGenesis,

                                                 inputType=INPUT_RESTART,
                                                 restartProt=protGenesisMin,

                                                  simulationType=SIMULATION_NMMD,
                                                  time_step=0.0005,
                                                  n_steps=1000,
                                                  eneout_period=100,
                                                  crdout_period=100,
                                                  nbupdate_period=10,
                                                  nm_number=6,
                                                  nm_mass=1.0,
                                                  inputModes=protNMA.outputModes,

                                                  implicitSolvent=IMPLICIT_SOLVENT_NONE,
                                                  electrostatics=ELECTROSTATICS_CUTOFF,
                                                  switch_dist=10.0,
                                                  cutoff_dist=12.0,
                                                  pairlist_dist=15.0,

                                                  ensemble=ENSEMBLE_NVT,
                                                  tpcontrol=TPCONTROL_LANGEVIN,
                                                  temperature=50.0,

                                                  boundary=BOUNDARY_NOBC,
                                                  EMfitChoice=EMFIT_IMAGES,
                                                  constantK="500",
                                                  emfit_sigma=2.0,
                                                  emfit_tolerance=0.1,
                                                  inputImage=target_images.outputImages,
                                                  pixel_size=2.0,
                                                  imageAngleShift=target_images._getExtraPath("GroundTruth.xmd"),

                                                  numberOfThreads=1,
                                                  numberOfMpi=NUMBER_OF_CPU,
                                                  )
            protGenesisFitNMMDImg.setObjLabel('NMMD Flexible Fitting Images')

            # Launch Fitting
            self.launchProtocol(protGenesisFitNMMDImg)

    # def test3_MDCHARMM(self):
    #     # Import PDB
    #     protPdbIonize = self.newProtocol(ProtImportPdb, inputPdbData=1,
    #                                    pdbFile=self.ds.getFile('4ake_solvate_pdb'))
    #     protPdbIonize.setObjLabel('Input PDB (4AKE solvated with water & ions)')
    #     self.launchProtocol(protPdbIonize)
    #
    #     # Minimize energy
    #     protGenesisMin = self.newProtocol(ProtGenesis,
    #           inputPDB = protPdbIonize.outputPdb,
    #           forcefield = FORCEFIELD_CHARMM,
    #           inputPRM = self.ds.getFile('charmm_prm'),
    #           inputRTF = self.ds.getFile('charmm_top'),
    #           inputPSF = self.ds.getFile('4ake_solvate_psf'),
    #           inputSTR = self.ds.getFile('charmm_str'),
    #
    #           simulationType = SIMULATION_MIN,
    #           time_step = 0.002,
    #           n_steps = 100, # 2000
    #           eneout_period = 10,
    #           crdout_period = 10,
    #           nbupdate_period = 10,
    #
    #           electrostatics = ELECTROSTATICS_PME,
    #           switch_dist = 10.0,
    #           cutoff_dist = 12.0,
    #           pairlist_dist = 15.0,
    #
    #           boundary = BOUNDARY_PBC,
    #           box_size_x=84.99,
    #           box_size_y=102.98,
    #           box_size_z=99.25,
    #
    #           rigid_bond = True,
    #           fast_water = True,
    #           water_model = "TIP3",
    #
    #           numberOfThreads=NUMBER_OF_CPU,
    #       )
    #     protGenesisMin.setObjLabel("[GENESIS]\n Energy Minimization CHARMM Explicit solvent")
    #     # Launch minimisation
    #     self.launchProtocol(protGenesisMin)
    #
    #     # Get GENESIS log file
    #     output_prefix = protGenesisMin.getOutputPrefix()
    #     log_file = output_prefix + ".log"
    #
    #     # Get the potential energy from the log file
    #     potential_ene = readLogFile(log_file)["POTENTIAL_ENE"]
    #
    #     # Assert that the potential energy is decreasing
    #     print("\n\n//////////////////////////////////////////////")
    #     print(protGenesisMin.getObjLabel())
    #     print("Initial potential energy : %.2f kcal/mol" % potential_ene[0])
    #     print("Final potential energy : %.2f kcal/mol" % potential_ene[-1])
    #     print("//////////////////////////////////////////////\n\n")
    #
    #     assert (potential_ene[0] > potential_ene[-1])
    #
    #     protGenesisMDRun = self.newProtocol(ProtGenesis,
    #                 inputPDB=protGenesisMin.outputPDB,
    #                 forcefield=FORCEFIELD_CHARMM,
    #                 inputPRM=self.ds.getFile('charmm_prm'),
    #                 inputRTF=self.ds.getFile('charmm_top'),
    #                 inputPSF=self.ds.getFile('4ake_solvate_psf'),
    #                 inputSTR=self.ds.getFile('charmm_str'),
    #                 restartchoice=True,
    #                 inputRST=protGenesisMin.getOutputPrefix() + ".rst",
    #
    #                 integrator=INTEGRATOR_NMMD,
    #                 time_step=0.002,
    #                 n_steps=10,
    #                 eneout_period=10,
    #                 crdout_period=10,
    #                 nbupdate_period=10,
    #                 nm_number=6,
    #                 nm_mass=1.0,
    #
    #                 electrostatics=ELECTROSTATICS_PME,
    #                 switch_dist=10.0,
    #                 cutoff_dist=12.0,
    #                 pairlist_dist=15.0,
    #
    #                 ensemble=ENSEMBLE_NPT,
    #                 tpcontrol=TPCONTROL_LANGEVIN,
    #                 temperature=300.0,
    #                 pressure=1.0,
    #
    #                 boundary=BOUNDARY_PBC,
    #                 box_size_x=84.99,
    #                 box_size_y=102.98,
    #                 box_size_z=99.25,
    #
    #                 rigid_bond=True,
    #                 fast_water=True,
    #                 water_model="TIP3",
    #
    #                 EMfitChoice=EMFIT_VOLUMES,
    #                 constantK=10000,
    #                 emfit_sigma=2.0,
    #                 emfit_tolerance=0.1,
    #                 inputVolume=self.protImportVol.outputVolume,
    #                 voxel_size=2.0,
    #                 centerOrigin=True,
    #
    #                 numberOfThreads=NUMBER_OF_CPU,
    #               )
    #     protGenesisMDRun.setObjLabel("[GENESIS]\n MD simulation with CHARMM explicit solvent")
    #     # Launch Simulation
    #     self.launchProtocol(protGenesisMDRun)


# **************************************************************************
# * Authors:     Mohamad Harastani (mohamad.harastani@igbmc.fr)
# * IMPMC, Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pwem.protocols import ProtImportPdb
from pwem.tests.workflows import TestWorkflow
from pyworkflow.tests import setupTestProject, DataSet
from continuousflex.protocols import FlexProtSynthesizeSubtomo, FlexProtMissingWedgeRestoration, FlexProtVolumeDenoise


class BM4D_and_MWR(TestWorkflow):
    """ Test protocol for BM4D and MWR. """
    @classmethod
    def setUpClass(cls):
        # Create a new project
        setupTestProject(cls)
        cls.ds = DataSet.getDataSet('nma_V2.0')

    def test_BM4D(self):
        """ Run NMA simple workflow for both Atomic and Pseudoatoms. """
        # Import PDB
        protImportPdb = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                         pdbFile=self.ds.getFile('pdb'))
        protImportPdb.setObjLabel('AK.pdb')
        self.launchProtocol(protImportPdb)
        SNR = 0.1
        N = 2
        # Synthesize subtomograms
        protSynthesize = self.newProtocol(FlexProtSynthesizeSubtomo,
                                          confVar=0,
                                          numberOfVolumes=N,
                                          targetSNR=SNR,
                                          volumeSize=32,
                                          samplingRate=4.4,
                                          )
        protSynthesize.refAtomic.set(protImportPdb.outputPdb)
        protSynthesize.setObjLabel('subtomograms')
        self.launchProtocol(protSynthesize)

        # Missing wedge restoration
        protMWC = self.newProtocol(FlexProtMissingWedgeRestoration,
                                   T=5,
                                   )
        protMWC.inputVolumes.set(protSynthesize.outputVolumes)
        protMWC.setObjLabel('missing wedge restoration')
        self.launchProtocol(protMWC)

        # Volume denoising
        protDenoise = self.newProtocol(FlexProtVolumeDenoise)
        protDenoise.inputVolumes.set(protSynthesize.outputVolumes)
        protDenoise.setObjLabel('Bm4D volume denoising')
        self.launchProtocol(protDenoise)



# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pyworkflow.viewer import (ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO)
from continuousflex.protocols.protocol_genesis import *
from continuousflex.protocols.utilities.genesis_utilities import *
from .plotter import FlexPlotter
from pwem.viewers import VmdView, ChimeraView
from pyworkflow.utils import getListFromRangeString
import numpy as np
import os
import glob
from matplotlib.pyplot import cm


class GenesisViewer(ProtocolViewer):
    """ Visualization of results from the GENESIS protocol
    """
    _label = 'viewer genesis'
    _targets = [ProtGenesis]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def _defineParams(self, form):
        form.addSection(label='Visualization')

        if self.protocol.getNumberOfSimulation() >1:
            form.addParam('fitRange', params.NumericRangeParam,
                          label="Simulation selection",
                          default="1-%i"%self.protocol.getNumberOfSimulation(),
                          important = True,
                          help=' Select the simulation to display. Examples:'
                               ' "1,3-5" -> [1,3,4,5]'
                               ' "1, 2, 4" -> [1,2,4]')
        if self.protocol.simulationType.get() == SIMULATION_REMD\
                or self.protocol.simulationType.get() == SIMULATION_RENMMD:
            form.addParam('replicaRange', params.NumericRangeParam,
                          label="Replica selection",
                          default="1-%i"%self.protocol.nreplica.get(),
                          help=' Select the replicas to display. Examples:'
                               ' "1,3-5" -> [1,3,4,5]'
                               ' "1, 2, 4" -> [1,2,4]')

        form.addParam('compareToPDB', params.BooleanParam, default=False,
                        label="Compare to external PDB",
                        help='TODO')

        group = form.addGroup('External PDB',condition= "compareToPDB")
        group.addParam('targetPDB', params.PathParam, default=None,
                        label="Target PDB (s)", important=True,
                        help=' Target PDBs to compute RMSD against. Atom mathcing is performed between '
                             ' the output PDBs and the target PDBs. Use the file pattern as file location with /*.pdb',
                       condition= "compareToPDB")
        group.addParam('referencePDB', params.PathParam, default="",
                        label="Intial PDB",
                        help='Atom matching will ignore the output PDB and will use the initial PDB instead.',
                        expertLevel=params.LEVEL_ADVANCED,condition= "compareToPDB")

        group.addParam('alignTarget', params.BooleanParam, default=False,
                        label="Align Target PDB",
                        help='TODO',condition= "compareToPDB")

        group = form.addGroup('Chimera 3D view')
        group.addParam('displayChimera', params.LabelParam,
                      label='Display results in Chimera',
                      help='Show initial and final structures in Chimera')

        group = form.addGroup('VMD trajectory view')
        group.addParam('displayTrajVMD', params.LabelParam,
                      label='Display trajecory in VMD',
                      help='TODO')

        group = form.addGroup('Energy Analysis')
        group.addParam('displayEnergy', params.LabelParam,
                      label='Display Potential Energy',
                      help='Show time series of the potentials used in MD simulation/Minimization')

        group = form.addGroup('RMSD analysis', condition= "compareToPDB")
        group.addParam('displayRMSDts', params.LabelParam,
                      label='Display RMSD time series',
                      help='TODO',condition= "compareToPDB")

        group.addParam('displayRMSD', params.LabelParam,
                      label='Display final RMSD',
                      help='TODO',condition= "compareToPDB")


        if self.protocol.EMfitChoice.get() != EMFIT_NONE:
            group = form.addGroup('Cryo EM fitting')
            group.addParam('displayCC', params.LabelParam,
                          label='Display Correlation Coefficient',
                          help='Show C.C. time series during the simulation')

    def _getVisualizeDict(self):
        return {
            'displayChimera': self._plotChimera,
            'displayEnergy': self._plotEnergy,
            'displayCC': self._plotCC,
            'displayRMSDts': self._plotRMSDts,
            'displayRMSD': self._plotRMSD,
            'displayTrajVMD': self._plotTrajVMD,
                }

    def _plotChimera(self, paramName):
        tmpChimeraFile = self.protocol._getExtraPath("chimera.cxc")
        index = self.getSimulationList()[0]

        with open(tmpChimeraFile, "w") as f:
            f.write("open %s.pdb \n"% os.path.abspath(self.protocol.getInputPDBprefix(index)))
            f.write("color #1 magenta \n" )
            count = 1

            outpdbfile = self.getOutputPrefixAll(index)[0] +".pdb"
            if os.path.exists(outpdbfile):
                if os.path.getsize(outpdbfile) != 0:
                    f.write("open %s \n"% os.path.abspath(outpdbfile))
                    count+=1
                    f.write("color #%s lime \n"%count)

            if self.compareToPDB.get():
                f.write("open %s \n" % os.path.abspath(self.getTargetPDB(index)))
                count+=1
                f.write("color #%s orange \n"%count)

            if self.protocol.EMfitChoice.get() == EMFIT_VOLUMES:
                f.write("open %s.mrc \n" % os.path.abspath(self.protocol.getInputEMprefix(index)))
                count+=1
                f.write("volume #%i transparency 0.5\n"%count)

            f.write("hide atoms \n")
            f.write("show cartoons \n")
            f.write("lighting soft \n")

        cv = ChimeraView(tmpChimeraFile)
        cv.show()

    def _plotTrajVMD(self, paramName):
        tmpVmdFile = self.protocol._getExtraPath("vmd.tcl")
        index = self.getSimulationList()[0]
        with open(tmpVmdFile, "w") as f:
            f.write("mol new %s.pdb waitfor all\n" % self.protocol.getInputPDBprefix(index))
            f.write("mol addfile %s.dcd waitfor all\n" % self.getOutputPrefixAll(index)[0])

            if self.protocol.forcefield.get() == FORCEFIELD_CHARMM:
                f.write("mol modstyle 0 0 NewCartoon \n")
            else:
                f.write("mol modstyle 0 0 Tube \n")
            f.write("mol modcolor 0 0 Molecule\n")

            if self.protocol.EMfitChoice.get() == EMFIT_VOLUMES:
                f.write("mol addfile %s.mrc waitfor all\n" % self.protocol.getInputEMprefix(index))
                f.write("mol addrep 0 \n")
                f.write("mol modstyle 1 0 Isosurface 0.5 0 0 0 1 1  \n")
                f.write("mol modmaterial 1 0 Transparent \n")

            if self.compareToPDB.get():
                targetFile = self.getTargetPDB(index)
                f.write("set nf [molinfo top get numframes]\n")
                f.write("mol new %s waitfor all\n" %targetFile)
                f.write( "for {set i 1 } {$i < $nf} {incr i} {\n")
                f.write( "animate dup frame 0 1\n")
                f.write("}\n")
                if self.protocol.forcefield.get() == FORCEFIELD_CHARMM:
                    f.write("mol modstyle 0 1 NewCartoon \n")
                else:
                    f.write("mol modstyle 0 1 Tube \n")
                f.write("mol modcolor 0 1 Molecule\n")
            f.write("animate style Loop\n")
            f.write("display projection Orthographic\n")

        vmdviewer = VmdView(" -e " + tmpVmdFile)
        vmdviewer.show()

    def _plotEnergy(self, paramName):
        self._plotEnergyTotal()
        self._plotEnergyDetail()

    def _plotEnergyTotal(self):
        ene_default = ["TOTAL_ENE", "POTENTIAL_ENE", "KINETIC_ENE"]

        ene = {}
        for i in self.getSimulationList():
            outputPrefix = self.getOutputPrefixAll(i)
            for j in outputPrefix:
                log_file = readLogFile(j + ".log")
                for e in ene_default:
                    if e in log_file:
                        if e in ene :
                            ene[e].append(log_file[e])
                        else:
                            ene[e] = [log_file[e]]
        enelist =[]
        labels=[]
        for e in ene :
            labels.append(e)
            enelist.append(ene[e])


        self.genesisPlotter(title="Energy (kcal/mol)", data=enelist, ndata=len(enelist),
                            nrep=len(enelist[0]), labels=labels)

    def _plotEnergyDetail(self):
        ene_default = ["BOND", "ANGLE", "UREY-BRADLEY", "DIHEDRAL", "IMPROPER", "CMAP", "VDWAALS", "ELECT", "NATIVE_CONTACT",
               "NON-NATIVE_CONT", "RESTRAINT_TOTAL"]

        ene = {}
        for i in self.getSimulationList():
            outputPrefix = self.getOutputPrefixAll(i)
            for j in outputPrefix:
                log_file = readLogFile(j+".log")
                for e in ene_default:
                    if e in log_file:
                        if e in ene :
                            ene[e].append(log_file[e])
                        else:
                            ene[e] = [log_file[e]]
        enelist =[]
        labels=[]
        for e in ene :
            labels.append(e)
            enelist.append(ene[e])

        self.genesisPlotter(title="Energy (kcal/mol)", data=enelist, ndata=len(enelist),
                            nrep=len(enelist[0]), labels=labels)

    def _plotCC(self, paramName):
        # Get CC list
        cc = []
        labels=[]
        simlist = self.getSimulationList()
        for i in simlist:
            outputPrefix = self.getOutputPrefixAll(i)
            cc_rep = []
            if len(simlist) == 1:
                labels.append("CC")
            else:
                labels.append("CC %s" % str(i + 1))
            for j in outputPrefix:
                log_file = readLogFile(j + ".log")
                if 'RESTR_CVS001' in log_file:
                    cc_rep.append(log_file['RESTR_CVS001'])
                else:
                    raise RuntimeError("CC not present in the log file")
            cc.append(cc_rep)

        self.genesisPlotter(title="CC", data=cc, ndata=len(simlist),
                            nrep=len(self.getOutputPrefixAll()), labels=labels)


    def genesisPlotter(self, title, data, ndata, nrep, labels):
        plotter = FlexPlotter()
        ax = plotter.createSubPlot(title, "", title)
        nmax= 10
        colors = [cm.get_cmap("tab10", 10)(i) for i in range(nmax) ]

        for i in range(ndata):
            if ndata <= nmax and nrep > 1:
                try:
                    meandata = np.mean(data[i], axis=0)
                    stddata = np.std(data[i], axis=0)
                    x = self.getTimePeriod(len(meandata))
                    ax.errorbar(x=x, y=meandata, yerr=stddata,
                                capthick=1.7, capsize=5, elinewidth=1.7, color=colors[i] if ndata!= 1 else "black",
                                errorevery=np.max([len(meandata) // 10, 1]), label="%s Average" % labels[i])
                except TypeError:
                    x = self.getTimePeriod(len(data[i][0]))
                    ax.plot(x, data[i][0], color=colors[i], label=labels[i])

            for j in range(nrep):
                x = self.getTimePeriod(len(data[i][j]))
                if 1 < nrep <= nmax:
                    if ndata == 1 :
                        ax.plot(x, data[i][j], color= colors[j], alpha=0.5, label="#%i"%(j+1))
                    # else:
                    #     ax.plot(x, data[i][j], color= colors[i], alpha=0.5)
                if nrep == 1 and ndata <= 10:
                    ax.plot(x, data[i][j], color= colors[i],label=labels[i])
        if ndata > nmax :
            try:
                meandata = np.mean(data, axis=(0,1))
                stddata = np.std(data, axis=(0,1))
                x = self.getTimePeriod(len(meandata))
                ax.errorbar(x=x, y=meandata, yerr=stddata,
                            capthick=1.7, capsize=5, elinewidth=1.7, color="black",
                            errorevery=np.max([len(meandata) // 10, 1]), label="Global Average")
            except TypeError:
                x = self.getTimePeriod(len(data[0][0]))
                ax.plot(x, data[0][0], color=colors[0], label=labels[0])
        xlabel = "Number of iterations" if self.protocol.simulationType.get() == SIMULATION_MIN else "Time (ps)"
        ax.set_xlabel(xlabel)
        plotter.legend()
        plotter.show()

    def _plotRMSDts(self, paramName):

        # Get matching atoms
        if self.referencePDB.get() != "":
            ref_pdb = ContinuousFlexPDBHandler(self.referencePDB.get())
        else:
            ref_pdb = ContinuousFlexPDBHandler(self.protocol.getInputPDBprefix()+".pdb")
        target_pdb = ContinuousFlexPDBHandler(self.getTargetPDB())
        idx_matchin_atoms = ref_pdb.matchPDBatoms(reference_pdb=target_pdb, ca_only=True)

        # Get RMSD list
        rmsd = []
        labels=[]
        simlist = self.getSimulationList()
        for i in simlist:
            outputPrefix = self.getOutputPrefixAll(i)
            if len(simlist) == 1:
                labels.append("RMSD")
            else:
                labels.append("RMSD %s"%str(i+1))
            rmsd_rep=[]
            for outprf in outputPrefix:
                rmsd_curr = []

                inputPDB = ContinuousFlexPDBHandler(self.protocol.getInputPDBprefix(i)+".pdb")
                targetPDB = ContinuousFlexPDBHandler(self.getTargetPDB(i))
                rmsd_curr.append(inputPDB.getRMSD(reference_pdb=targetPDB, align=self.alignTarget.get(), idx_matching_atoms=idx_matchin_atoms))
                coord_arr = dcd2numpyArr(outprf + ".dcd")
                for i in range(len(coord_arr)):
                    inputPDB.coords[:, :] = coord_arr[i]
                    rmsd_curr.append(inputPDB.getRMSD(reference_pdb=targetPDB, align=self.alignTarget.get(), idx_matching_atoms=idx_matchin_atoms))

                rmsd_rep.append(rmsd_curr)
            rmsd.append(rmsd_rep)

        self.genesisPlotter(title="RMSD ($\AA$)", data=rmsd, ndata=len(simlist),
                            nrep=len(self.getOutputPrefixAll()), labels=labels)


    def _plotRMSD(self, paramName):
        plotter = FlexPlotter()
        ax = plotter.createSubPlot("RMSD ($\AA$)", "# Simulation", "RMSD ($\AA$)")

        # Get RMSD list
        initial_mols = []
        final_mols = []
        target_mols = []
        for i in self.getSimulationList():
            inputPDB = self.protocol.getInputPDBprefix(i)+".pdb"
            targetPDB = self.getTargetPDB(i)
            outputPrefs = self.getOutputPrefixAll(i)
            target_mols.append(ContinuousFlexPDBHandler(targetPDB))
            initial_mols.append(ContinuousFlexPDBHandler(inputPDB))
            for outputPrefix in outputPrefs:
                outputPDB = outputPrefix +".pdb"
                final_mols.append(ContinuousFlexPDBHandler(outputPDB))

        if self.referencePDB.get() != "":
            ref_mol = ContinuousFlexPDBHandler(self.referencePDB.get())
        else:
            ref_mol = initial_mols[0]
        idx_match = ref_mol.matchPDBatoms(reference_pdb=target_mols[0],ca_only=True)
        rmsdi=[]
        rmsdf=[]
        for i in range(len(self.getSimulationList())):
            for j in range(len(outputPrefs)):
                rmsdi.append(initial_mols[i].getRMSD(reference_pdb=target_mols[i],
                                                     idx_matching_atoms=idx_match,
                                                     align=self.alignTarget.get()))
                rmsdf.append(final_mols[i*len(outputPrefs) + j].getRMSD(reference_pdb=target_mols[i],
                                                    idx_matching_atoms=idx_match, align=self.alignTarget.get()))

        ax.plot(rmsdf, "o", color="tab:blue", label="Final RMSD", markeredgecolor='black')
        ax.plot(rmsdi, "o", color="tab:green", label="Initial RMSD", markeredgecolor='black')
        plotter.legend()
        plotter.show()

    def getSimulationList(self):
        if self.protocol.getNumberOfSimulation() > 1:
            return np.array(getListFromRangeString(self.fitRange.get())) -1
        else:
            return np.array([0])

    def getTargetPDB(self, index=0):
        targetPDBlist = [f for f in glob.glob(self.targetPDB.get())]
        targetPDBlist.sort()
        if index < len(targetPDBlist):
            return targetPDBlist[index]
        else:
            return targetPDBlist[0]

    def getOutputPrefixAll(self, index=0):
        outPrf = np.array(self.protocol.getOutputPrefixAll(index))
        if self.protocol.simulationType.get() == SIMULATION_REMD \
                or self.protocol.simulationType.get() == SIMULATION_RENMMD:
            return outPrf[np.array(getListFromRangeString(self.replicaRange.get())) - 1]
        else:
            return outPrf

    def getTimePeriod(self, length):
        if self.protocol.simulationType.get() == SIMULATION_MIN:
            timestep = 1.0
        else:
            timestep = float(self.protocol.time_step.get())
        eneperiod = int(self.protocol.eneout_period.get())

        return np.arange(length) * eneperiod * timestep




# This function converts a Dynamo table into xmipp (scipion) metadata file
# given a metadata input containing a list of the subtomograms
from pwem.emlib import metadata as md
import numpy as np
from math import sin, cos, radians
from xmippLib import Euler_matrix2angles
import pandas as pd
import math


def dynamo_mat(tdrot, tilt, narot, shiftx, shifty, shiftz):
    tdrot = radians(tdrot)
    tilt = radians(tilt)
    narot = radians(narot)
    cotd = cos(tdrot)
    sitd = sin(tdrot)
    coti = cos(tilt)
    siti = sin(tilt)
    cona = cos(narot)
    sina = sin(narot)
    m = np.zeros([4,4])
    m[0,0] = cotd * cona - sitd * coti * sina
    m[1,0] = - cona * sitd - cotd * coti * sina
    m[2,0] = sina * siti
    m[0,1] = cotd * sina + cona * sitd * coti
    m[1,1] = cotd * cona * coti - sitd * sina
    m[2,1] = -cona * siti
    m[0,2] = sitd * siti
    m[1,2] = cotd * siti
    m[2,2] = coti
    # The 4th column
    m[0,3] = shiftx
    m[1,3] = shifty
    m[2,3] = shiftz
    m[3,3] = 1

    return m


def matrix2eulerAngles(A):
    abs_sb = np.sqrt(A[0, 2] * A[0, 2] + A[1, 2] * A[1, 2])
    if (abs_sb > 16 * np.exp(-5)):
        gamma = math.atan2(A[1, 2], -A[0, 2])
        alpha = math.atan2(A[2, 1], A[2, 0])
        if (abs(np.sin(gamma)) < np.exp(-5)):
            sign_sb = np.sign(-A[0, 2] / np.cos(gamma))
        else:
            if np.sin(gamma) > 0:
                sign_sb = np.sign(A[1, 2])
            else:
                sign_sb = -np.sign(A[1, 2])
        beta = math.atan2(sign_sb * abs_sb, A[2, 2])
    else:
        if (np.sign(A[2, 2]) > 0):
            alpha = 0
            beta = 0
            gamma = math.atan2(-A[1, 0], A[0, 0])
        else:
            alpha = 0
            beta = np.pi
            gamma = math.atan2(A[1, 0], -A[0, 0])
    gamma = np.rad2deg(gamma)
    beta = np.rad2deg(beta)
    alpha = np.rad2deg(alpha)
    return alpha, beta, gamma, A[0, 3], A[1, 3], A[2, 3]


def tbl2metadata(table, mdfi, mdfo):
    tbl = pd.read_csv(table, delimiter=' ', header=None)
    x = tbl[:][3]
    y = tbl[:][4]
    z = tbl[:][5]
    tdrot = tbl[:][6]
    tiltd = tbl[:][7]
    narot = tbl[:][8]
    # change the angles from Dynamo convention to xmipp convention
    rot = []
    tilt = []
    psi = []
    shiftx = []
    shifty = []
    shiftz = []
    for i in range(tbl.shape[0]):
        TransMat = dynamo_mat(tdrot[i], tiltd[i], narot[i], x[i], y[i], z[i])
        rot_i, tilt_i, psi_i, shiftx_i, shifty_i, shiftz_i = matrix2eulerAngles(TransMat)
        rot.append(rot_i)
        tilt.append(tilt_i)
        psi.append(psi_i)
        shiftx.append(shiftx_i)
        shifty.append(shifty_i)
        shiftz.append(shiftz_i)

    md_out = md.MetaData(mdfi)
    i = 0
    for objId in md_out:
        md_out.setValue(md.MDL_SHIFT_X,    shiftx[i],objId)
        md_out.setValue(md.MDL_SHIFT_Y,    shifty[i],objId)
        md_out.setValue(md.MDL_SHIFT_Z,    shiftz[i], objId)
        md_out.setValue(md.MDL_ANGLE_ROT,  rot[i], objId)
        md_out.setValue(md.MDL_ANGLE_TILT, tilt[i], objId)
        md_out.setValue(md.MDL_ANGLE_PSI,  psi[i], objId)
        md_out.setValue(md.MDL_ANGLE_Y,    0.0, objId)
        i += 1
    pass
    md_out.write(mdfo)



# By Remi Vuillemot

import numpy as np
from pyworkflow.utils import runCommand
import pwem.emlib.metadata as md
import re
import multiprocessing

NUMBER_OF_CPU = int(np.min([multiprocessing.cpu_count(),4]))

EMFIT_NONE = 0
EMFIT_VOLUMES = 1
EMFIT_IMAGES = 2

FORCEFIELD_CHARMM = 0
FORCEFIELD_AAGO = 1
FORCEFIELD_CAGO = 2

SIMULATION_MIN = 0
SIMULATION_MD = 1
SIMULATION_NMMD = 2
SIMULATION_REMD = 3
SIMULATION_RENMMD = 4

PROGRAM_ATDYN = 0
PROGRAM_SPDYN= 1

INTEGRATOR_VVERLET = 0
INTEGRATOR_LEAPFROG = 1

IMPLICIT_SOLVENT_GBSA = 0
IMPLICIT_SOLVENT_NONE = 1

TPCONTROL_NONE = 0
TPCONTROL_LANGEVIN = 1
TPCONTROL_BERENDSEN = 2
TPCONTROL_BUSSI = 3

ENSEMBLE_NVT = 0
ENSEMBLE_NVE = 1
ENSEMBLE_NPT = 2

BOUNDARY_NOBC = 0
BOUNDARY_PBC = 1

ELECTROSTATICS_PME = 0
ELECTROSTATICS_CUTOFF = 1

NUCLEIC_NO = 0
NUCLEIC_RNA =1
NUCLEIC_DNA = 2

RB_PROJMATCH = 0
RB_WAVELET = 1

INPUT_TOPOLOGY = 0
INPUT_RESTART = 1
INPUT_NEW_SIM = 2

PROJECTION_ANGLE_SAME=0
PROJECTION_ANGLE_XMIPP=1
PROJECTION_ANGLE_IMAGE=2

def save_dcd(mol, coords_list, prefix):
    print("> Saving DCD trajectory ...")
    n_frames = len(coords_list)

    # saving PDBs
    mol = mol.copy()
    for i in range(n_frames):
        mol.coords = coords_list[i]
        mol.write_pdb("%s_frame%i.pdb" % (prefix, i))

    # VMD command
    with open(prefix+"_cmd.tcl", "w") as f :
        f.write("mol new %s_frame0.pdb\n" % prefix)
        for i in range(1,n_frames):
            f.write("mol addfile %s_frame%i.pdb\n" % (prefix, i))
        f.write("animate write dcd %s.dcd\n" % prefix)
        f.write('exit\n')

    # Running VMD
    runCommand("vmd -dispdev text -e %s_cmd.tcl" % prefix)

    # Cleaning
    for i in range(n_frames):
        runCommand("rm -f %s_frame%i.pdb\n" % (prefix, i))
    runCommand("rm -f %s_cmd.tcl" % prefix)
    print("\t Done \n")


def lastPDBFromDCD(inputPDB,inputDCD,  outputPDB):

    # EXTRACT PDB from dcd file
    with open("%s_tmp_dcd2pdb.tcl" % outputPDB, "w") as f:
        s = ""
        s += "mol load pdb %s dcd %s\n" % (inputPDB, inputDCD)
        s += "set nf [molinfo top get numframes]\n"
        s += "[atomselect top all frame [expr $nf - 1]] writepdb %s\n" % outputPDB
        s += "exit\n"
        f.write(s)
    runCommand("vmd -dispdev text -e %s_tmp_dcd2pdb.tcl" % outputPDB)

    # CLEAN TMP FILES
    runCommand("rm -f %s_tmp_dcd2pdb.tcl" % (outputPDB))

def buildParallelScript(commands,numberOfThreads=1,  raiseError=True):
    """
    :param list commands: list of commands to run in parallel
    :param numberOfThreads: Number of openMP threads
    :param raiseError: raise error if fails
    :return None:
    """

    py_script =\
        """
from mpi4py import MPI
import sys
import os
from subprocess import Popen
comm = MPI.COMM_WORLD
rank = comm.Get_rank()

env = os.environ
env["OMP_NUM_THREADS"] = str(%i)

        """%numberOfThreads
    for i in range(len(commands)):
        py_script +=\
            """
if rank == %i:
    p = Popen("%s", shell=True, stdout=sys.stdout, stderr = sys.stderr, env=env)
    exitcode = p.wait()
    if exitcode != 0:
        err_msg = "Command returned with errors : %s"
        if %s :
            raise RuntimeError(err_msg)
        else:
            print(err_msg)
            """ % (i, commands[i], commands[i], "True" if raiseError else "False")


    py_script +=\
    """
exit(0)
    """
    return py_script





def pdb2vol(inputPDB, outputVol, sampling_rate, image_size):
    """
    Create a density volume from a pdb
    :param str inputPDB: input pdb file name
    :param str outputVol: output vol file name
    :param float sampling_rate: Sampling rate
    :param int image_size: Size of the output volume
    :return str: the Xmipp command to run
    """
    cmd = "xmipp_volume_from_pdb"
    args = "-i %s  -o %s --sampling %f --size %i %i %i --centerPDB"%\
           (inputPDB, outputVol,sampling_rate,image_size,image_size,image_size)
    return cmd+ " "+ args

def projectVol(inputVol, outputProj, expImage, sampling_rate=5.0, angular_distance=-1, compute_neighbors=True):
    """
    Create a set of projections from an input volume
    :param str inputVol: Input volume file name
    :param str outputProj: Output set of proj file name
    :param str expImage: Experimental image to project in the neighborhood
    :param float sampling_rate: Samplign rate
    :param float angular_distance: Do not search a distance larger than...
    :param bool compute_neighbors: Compute projection nearby the experimental image
    :return str: the Xmipp command to run
    """
    cmd = "xmipp_angular_project_library"
    args = "-i %s.vol -o %s.stk --sampling_rate %f " % (inputVol, outputProj, sampling_rate)
    if compute_neighbors :
        args +="--compute_neighbors --angular_distance %f " % angular_distance
        args += "--experimental_images %s "%expImage
        if angular_distance != -1 :
            args += "--near_exp_data"
    return cmd+ " "+ args

def projectMatch(inputImage, inputProj, outputMeta):
    """
    Projection matching of an input experimental image with a set of projections
    :param str inputImage: File name of the input experimental image
    :param str inputProj: File name of the input set of projections
    :param str outputMeta: File name of the output Xmipp metadata file with the angles of the matching
    :return str: the Xmipp command to run
    """
    cmd = "xmipp_angular_projection_matching "
    args= "-i %s -o %s --ref %s.stk "%(inputImage, outputMeta, inputProj)
    args +="--search5d_shift 10.0 --search5d_step 1.0"
    return cmd + " "+ args

def waveletAssignement(inputImage, inputProj, outputMeta):
    """
    Make a discrete angular assignment of angles from a set of projections
    :param str inputImage: File name of input experimental image
    :param str inputProj: File name of the input set of projections
    :param str outputMeta: File name of the output Xmipp metadata file with the angles assigned
    :return str: the Xmipp command to run
    """
    cmd = "xmipp_angular_discrete_assign "
    args= "-i %s -o %s --ref %s.doc "%(inputImage, outputMeta, inputProj)
    args +="--psi_step 5.0 --max_shift_change 10.0 --search5D"
    return cmd + " "+ args

def continuousAssign(inputMeta, inputVol, outputMeta):
    """
    Make a continuous angular assignment of angles from a volume
    :param str inputMeta: File name of input Xmipp metadata file with angles to assign
    :param str inputVol: File name of the input Volume
    :param str outputMeta: File name of the output Xmipp metadata file with the angles
    :return str: the Xmipp command to run
    """
    cmd = "xmipp_angular_continuous_assign "
    args= "-i %s -o %s --ref %s.vol "%(inputMeta, outputMeta, inputVol)
    return cmd + " "+ args

def flipAngles(inputMeta, outputMeta):
    """
    Flip angles from Xmipp representation to Euler angles
    :param str inputMeta : File name of input Xmipp metadata file containing the angles to flip
    :param str outputMeta: file name of the output Xmipp matadata file
    :return None:
    """
    Md1 = md.MetaData(inputMeta)
    flip = Md1.getValue(md.MDL_FLIP, 1)
    tilt1 = Md1.getValue(md.MDL_ANGLE_TILT, 1)
    psi1 = Md1.getValue(md.MDL_ANGLE_PSI, 1)
    x1 = Md1.getValue(md.MDL_SHIFT_X, 1)
    if flip:
        Md1.setValue(md.MDL_SHIFT_X, -x1, 1)
        Md1.setValue(md.MDL_ANGLE_TILT, tilt1 + 180, 1)
        Md1.setValue(md.MDL_ANGLE_PSI, -psi1, 1)
    Md1.write(outputMeta)

# def getAngularDist(md1, md2, idx1=1, idx2=1):
#     rot1        = md1.getValue(md.MDL_ANGLE_ROT, int(idx1))
#     tilt1       = md1.getValue(md.MDL_ANGLE_TILT, int(idx1))
#     psi1        = md1.getValue(md.MDL_ANGLE_PSI, int(idx1))
#     rot2        = md2.getValue(md.MDL_ANGLE_ROT, int(idx2))
#     tilt2       = md2.getValue(md.MDL_ANGLE_TILT, int(idx2))
#     psi2        = md2.getValue(md.MDL_ANGLE_PSI, int(idx2))
#
#     return  SymList.computeDistanceAngles(SymList(), rot1, tilt1, psi1, rot2, tilt2, psi2, False, True, False)
#
#
# def getShiftDist(md1, md2, idx1=1, idx2=1):
#     shiftx1 = md1.getValue(md.MDL_SHIFT_X, int(idx1))
#     shifty1 = md1.getValue(md.MDL_SHIFT_Y, int(idx1))
#     shiftx2 = md2.getValue(md.MDL_SHIFT_X, int(idx2))
#     shifty2 = md2.getValue(md.MDL_SHIFT_Y, int(idx2))
#     return np.linalg.norm(np.array([shiftx1, shifty1, 0.0]) - np.array([shiftx2, shifty2, 0.0]))

def getAngularShiftDist(angle1MetaFile, angle2MetaData, angle2Idx, tmpPrefix, symmetry):

    mdImgTmp = md.MetaData()
    mdImgTmp.addObject()
    mdImgTmp.setValue(md.MDL_ANGLE_ROT, angle2MetaData.getValue(md.MDL_ANGLE_ROT, angle2Idx), 1)
    mdImgTmp.setValue(md.MDL_ANGLE_TILT,angle2MetaData.getValue(md.MDL_ANGLE_TILT,angle2Idx), 1)
    mdImgTmp.setValue(md.MDL_ANGLE_PSI, angle2MetaData.getValue(md.MDL_ANGLE_PSI, angle2Idx), 1)
    mdImgTmp.setValue(md.MDL_SHIFT_X,   angle2MetaData.getValue(md.MDL_SHIFT_X,   angle2Idx), 1)
    mdImgTmp.setValue(md.MDL_SHIFT_Y,   angle2MetaData.getValue(md.MDL_SHIFT_Y,   angle2Idx), 1)
    mdImgTmp.write(tmpPrefix + ".xmd")

    cmd = "xmipp_angular_distance --ang1 %s --ang2 %s.xmd --oroot %sDist --sym %s --check_mirrors > %s.log" % \
          (angle1MetaFile, tmpPrefix, tmpPrefix, symmetry, tmpPrefix)
    runCommand(cmd)
    with open(tmpPrefix + ".log", "r") as f:
        for line in f:
            if "angular" in line:
                angDist = float(re.findall("\d+\.\d+", line)[0])
            if "shift" in line:
                shftDist = float(re.findall("\d+\.\d+", line)[0])
    return angDist, shftDist


def readLogFile(log_file):
    with open(log_file,"r") as file:
        header = None
        dic = {}
        for line in file:
            if line.startswith("INFO:"):
                if header is None:
                    header = line.split()
                    for i in range(1,len(header)):
                        dic[header[i]] = []
                else:
                    splitline = line.split()
                    if len(splitline) == len(header):
                        for i in range(1,len(header)):
                            try :
                                dic[header[i]].append(float(splitline[i]))
                            except ValueError:
                                pass

    return dic

def dcd2numpyArr(filename):
    print("> Reading dcd file %s"%filename)
    BYTESIZE = 4
    with open(filename, 'rb') as f:

        # Header
        # ---------------- INIT

        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        crd_type = f.read(BYTESIZE).decode('ascii')
        nframe = int.from_bytes((f.read(BYTESIZE)), "little")
        start_frame = int.from_bytes((f.read(BYTESIZE)), "little")
        len_frame = int.from_bytes((f.read(BYTESIZE)), "little")
        len_total = int.from_bytes((f.read(BYTESIZE)), "little")
        for i in range(5):
            f.read(BYTESIZE)
        time_step = np.frombuffer(f.read(BYTESIZE), dtype=np.float32)
        for i in range(9):
            f.read(BYTESIZE)
        charmm_version = int.from_bytes((f.read(BYTESIZE)), "little")

        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ---------------- TITLE
        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        ntitle = int.from_bytes((f.read(BYTESIZE)), "little")
        tilte_rd = f.read(BYTESIZE*20 * ntitle)
        try :
            title = tilte_rd.encode("ascii")
        except AttributeError:
            title = str(tilte_rd)
        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ---------------- NATOM
        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        natom = int.from_bytes((f.read(BYTESIZE)), "little")
        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ----------------- DCD COORD
        dcd_arr =  np.zeros((nframe, natom, 3), dtype=np.float32)
        for i in range(nframe):
            for j in range(3):

                start_size = int.from_bytes((f.read(BYTESIZE)), "little")
                while (start_size != BYTESIZE * natom and start_size != 0):
                    # print("\n-- UNKNOWN %s -- " % start_size)

                    f.read(start_size)
                    end_size = int.from_bytes((f.read(BYTESIZE)), "little")
                    if end_size != start_size:
                        raise RuntimeError("Can not read dcd file")
                    start_size = int.from_bytes((f.read(BYTESIZE)), "little")

                bin_arr = f.read(BYTESIZE * natom)
                if len(bin_arr) == BYTESIZE * natom:
                    dcd_arr[i, :, j] = np.frombuffer(bin_arr, dtype=np.float32)
                else:
                    break
                end_size = int.from_bytes((f.read(BYTESIZE)), "little")
                if end_size != start_size:
                    if i>1:
                        break
                    else:
                        # pass
                        raise RuntimeError("Can not read dcd file %i %i " % (start_size, end_size))

        print("\t -- Summary of DCD file -- ")
        print("\t\t crd_type  : %s"%crd_type)
        print("\t\t nframe  : %s"%nframe)
        print("\t\t len_frame  : %s"%len_frame)
        print("\t\t len_total  : %s"%len_total)
        print("\t\t time_step  : %s"%time_step)
        print("\t\t charmm_version  : %s"%charmm_version)
        print("\t\t title  : %s"%title)
        print("\t\t natom  : %s"%natom)
    print("\t Done \n")

    return dcd_arr


def numpyArr2dcd(arr, filename, start_frame=1, len_frame=1, time_step=1.0, title=None):
    print("> Wrinting dcd file %s"%filename)
    BYTESIZE = 4
    nframe, natom, _ = arr.shape
    len_total=nframe*len_frame
    charmm_version=24
    if title is None:
        title = "DCD file generated by Continuous Flex plugin"
    ntitle = (len(title)//(20*BYTESIZE)) + 1
    with open(filename, 'wb') as f:
        zeroByte = int.to_bytes(0, BYTESIZE, "little")

        # Header
        # ---------------- INIT
        f.write(int.to_bytes(21*BYTESIZE ,BYTESIZE, "little"))
        f.write(b'CORD')
        f.write(int.to_bytes(nframe, BYTESIZE, "little"))
        f.write(int.to_bytes(start_frame, BYTESIZE, "little"))
        f.write(int.to_bytes(len_frame, BYTESIZE, "little"))
        f.write(int.to_bytes(len_total, BYTESIZE, "little"))
        for i in range(5):
            f.write(zeroByte)
        f.write(np.float32(time_step).tobytes())
        for i in range(9):
            f.write(zeroByte)
        f.write(int.to_bytes(charmm_version, BYTESIZE, "little"))

        f.write(int.to_bytes(21*BYTESIZE,BYTESIZE, "little"))

        # ---------------- TITLE
        f.write(int.to_bytes((ntitle*20+1)*BYTESIZE ,BYTESIZE, "little"))
        f.write(int.to_bytes(ntitle ,BYTESIZE, "little"))
        f.write(title.ljust(20*BYTESIZE).encode("ascii"))
        f.write(int.to_bytes((ntitle*20+1)*BYTESIZE ,BYTESIZE, "little"))

        # ---------------- NATOM
        f.write(int.to_bytes(BYTESIZE ,BYTESIZE, "little"))
        f.write(int.to_bytes(natom ,BYTESIZE, "little"))
        f.write(int.to_bytes(BYTESIZE ,BYTESIZE, "little"))

        # ----------------- DCD COORD
        for i in range(nframe):
            for j in range(3):
                f.write(int.to_bytes(BYTESIZE*natom, BYTESIZE, "little"))
                f.write(np.float32(arr[i, :, j]).tobytes())
                f.write(int.to_bytes(BYTESIZE*natom, BYTESIZE, "little"))
    print("\t Done \n")

def existsCommand(name):
    from shutil import which
    return which(name) is not None
